"""Tools for generic VMC optimization of tensor networks.
"""

import array
import random

import numpy as np
import autoray as ar

from quimb import format_number_with_error


# --------------------------------------------------------------------------- #


def sample_bitstring_from_prob_ndarray(p, rng):
    flat_idx = rng.choice(np.arange(p.size), p=p.flat)
    return np.unravel_index(flat_idx, p.shape)


def shuffled(it):
    """Return a copy of ``it`` in random order.
    """
    it = list(it)
    random.shuffle(it)
    return it


class NoContext:
    """A convenience context manager that does nothing.
    """

    def __enter__(self):
        pass

    def __exit__(self, *_, **__):
        pass


class MovingStatistics:
    """Keep track of the windowed mean and estimated variance of a stream of
    values on the fly.
    """

    def __init__(self, window_size):
        self.window_size = window_size
        self.xs = []
        self.vs = []
        self._xsum = 0.0
        self._vsum = 0.0

    def update(self, x):

        # update mean
        self.xs.append(x)
        if len(self.xs) > self.window_size:
            xr = self.xs.pop(0)
        else:
            xr = 0.0
        self._xsum += (x - xr)

        # update approx variance
        v = (x - self.mean)**2
        self.vs.append(v)
        if len(self.vs) > self.window_size:
            vr = self.vs.pop(0)
        else:
            vr = 0.0
        self._vsum += (v - vr)

    @property
    def mean(self) :
        N = len(self.xs)
        if N == 0:
            return 0.0
        return self._xsum / N

    @property
    def var(self):
        N = len(self.xs)
        if N == 0:
            return 0.0
        return self._vsum / N

    @property
    def std(self):
        return self.var**0.5

    @property
    def err(self):
        N = len(self.xs)
        if N == 0:
            return 0.0
        return self.std / N**0.5


# --------------------------------------------------------------------------- #

class DenseSampler:
    """Sampler that explicitly constructs the full probability distribution.
    Useful for debugging small problems.
    """

    def __init__(self, psi=None, seed=None, **contract_opts):
        if psi is not None:
            self._set_psi(psi)
        contract_opts.setdefault('optimize', 'auto-hq')
        self.contract_opts = contract_opts
        self.rng = np.random.default_rng(seed)

    def _set_psi(self, psi):
        psi_dense = psi.contract(
            ..., output_inds=psi.site_inds, **self.contract_opts,
        ).data
        self.p = (abs(psi_dense.ravel())**2)
        self.p /= self.p.sum()
        self.sites = psi.sites
        self.shape = tuple(psi.ind_size(ix) for ix in psi.site_inds)
        self.flat_indexes = np.arange(self.p.size)

    def sample(self):
        flat_idx = self.rng.choice(self.flat_indexes, p=self.p)
        omega = self.p[flat_idx]
        config = np.unravel_index(flat_idx, self.shape)
        return dict(zip(self.sites, config)), omega

    def update(self, **kwargs):
        self._set_psi(kwargs['psi'])


class DirectTNSampler:
    """

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to sample from.
    sweeps : int, optional
        The number of sweeps to perform.
    max_group_size : int, optional
        The maximum number of sites to include in a single marginal.
    chi : int, optional
        The maximum bond dimension to use for compressed contraction.
    optimize : PathOptimizer, optional
        The path optimizer to use.
    optimize_share_path : bool, optional
        If ``True``, a single path will be used for all contractions regardless
        of which marginal (i.e. which indices are open) is begin computed.
    """

    def __init__(
        self,
        tn,
        sweeps=1,
        max_group_size=8,
        chi=None,
        optimize=None,
        optimize_share_path=False,
        seed=None,
        track=False,
    ):
        self.tn = tn.copy()

        self.ind2site = {}
        self.tid2ind = {}
        for site in self.tn.sites:
            ix = self.tn.site_ind(site)
            tid, = self.tn._get_tids_from_inds(ix)
            self.tid2ind[tid] = ix
            self.ind2site[ix] = site

        self.chi = chi
        self.sweeps = sweeps
        self.max_group_size = max_group_size

        self.optimize = optimize
        self.optimize_share_path = optimize_share_path
        self.groups = None
        self.tree = None
        self.path = None

        self.rng = np.random.default_rng(seed)

        self.track = track
        if self.track:
            self.omegas = []
            self.probs = []
        else:
            self.omegas = self.probs = None

    def plot(self,):
        from matplotlib import pyplot as plt
        fig, ax = plt.subplots(figsize=(4, 4))
        mins = min(self.omegas)
        maxs = max(self.omegas)
        ax.plot([mins, maxs], [mins, maxs], color='red')
        ax.scatter(self.probs, self.omegas, marker='.', alpha=0.5)
        ax.set_xlabel('$\pi(x)$')
        ax.set_ylabel('$\omega(x)$')
        ax.set_xscale('log')
        ax.set_yscale('log')
        ax.grid(True, c=(0.97 ,0.97, 0.97), which='major')
        ax.set_axisbelow(True)

    def calc_groups(self, **kwargs):
        """Calculate how to group the sites into marginals.
        """
        self.groups = self.tn.compute_hierarchical_grouping(
            max_group_size=self.max_group_size,
            tids=tuple(self.tid2ind),
            **kwargs,
        )

    def get_groups(self):
        if self.groups is None:
            self.calc_groups()
        return self.groups

    def calc_path(self):
        tn0 = self.tn.isel({ix: 0 for ix in self.ind2site})
        self.tree = tn0.contraction_tree(self.optimize)
        self.path = self.tree.get_path()

    def get_path(self):
        if self.path is None:
            self.calc_path()
        return self.path

    def get_optimize(self):
        if self.optimize_share_path:
            return self.get_path()
        else:
            return self.optimize

    def contract(self, tn, output_inds):
        if self.chi is None:
            return tn.contract(
                optimize=self.get_optimize(),
                output_inds=output_inds,
            )
        else:
            return tn.contract_compressed(
                max_bond=self.chi,
                optimize=self.get_optimize(),
                output_inds=output_inds,
                cutoff=0.0,
                compress_opts=dict(absorb='both'),
            )

    def sample(self):

        config = {}

        tnm = self.tn.copy()

        for tid, ix in self.tid2ind.items():
            t = tnm.tensor_map[tid]
            t.rand_reduce_(
                ix, rand_fn=lambda d: self.rng.choice([-1.0, 1.0], size=d)
            )

        tnm.apply_to_arrays(ar.lazy.array)
        with ar.lazy.shared_intermediates():
        # with NoContext():

            for _ in range(self.sweeps):

                # random.shuffle(self.groups)
                omega = 1.0

                for group in self.get_groups():
                    # get corresponding indices
                    inds = [self.tid2ind[tid] for tid in group]

                    # insert the orig tensors with output index
                    for tid in group:
                        t_full = self.tn.tensor_map[tid]
                        tnm.tensor_map[tid].modify(
                            data=ar.lazy.array(t_full.data),
                            inds=t_full.inds,
                        )

                    # contract the current conditional marginal
                    tg = self.contract(tnm, inds)

                    # convert into normalized prob and sample a config
                    prob_g = ar.do('abs', tg.data.compute())**2
                    prob_g /= ar.do('sum', prob_g)
                    config_g = sample_bitstring_from_prob_ndarray(
                        prob_g, self.rng
                    )
                    omega *= prob_g[config_g]

                    # re-project the tensors according to the sampled config
                    for tid, ix, bi in zip(group, inds, config_g):

                        # the 'measurement' for this tensor
                        # bi = int(bi)

                        # project tensor from full wavefunction
                        t_full = self.tn.tensor_map[tid]
                        tm = t_full.isel({ix: bi})
                        tnm.tensor_map[tid].modify(
                            data=ar.lazy.array(tm.data),
                            inds=tm.inds,
                        )

                        # update the bitstring
                        config[self.ind2site[ix]] = bi

        if self.track:
            self.omegas.append(omega)
            self.probs.append(abs(tg.data[config_g].compute())**2)

        # final chosen marginal is prob of whole config
        return config, omega # tg.data[config_g].compute()


def compute_amplitude(tn, config, chi, optimize):
    tni = tn.isel({tn.site_ind(site): v for site, v in config.items()})
    return tni.contract_compressed_(
        optimize=optimize,
        max_bond=chi,
        cutoff=0.0,
        compress_opts={'absorb': 'both'},
        inplace=True,
    )


def compute_amplitudes(tn, configs, chi, optimize):
    with ar.lazy.shared_intermediates():
        tnlz = tn.copy()
        tnlz.apply_to_arrays(ar.lazy.array)

        amps = []
        for config in configs:
            amps.append(compute_amplitude(tnlz, config, chi, optimize))

    amps = ar.do('stack', amps)
    return amps.compute()


def compute_local_energy(ham, tn, config, chi, optimize):
    """
    """
    c_configs, c_coeffs = ham.config_coupling(config)
    amps = compute_amplitudes(tn, [config] + c_configs, chi, optimize)
    c_coeffs = ar.do('array', c_coeffs, like=amps)
    return ar.do('sum', amps[1:] * c_coeffs) / amps[0]


def draw_config(edges, config):
    import networkx as nx
    G = nx.Graph(edges)
    pos = nx.kamada_kawai_layout(G)
    nx.draw(G, node_color=[config[node] for node in G.nodes], pos=pos)


class ClusterSampler:

    def __init__(
        self,
        psi=None,
        max_distance=1,
        use_gauges=True,
        seed=None,
        contract_opts=None,
    ):
        self.rng = np.random.default_rng(seed)
        self.use_gauges = use_gauges
        self.max_distance = max_distance
        self.contract_opts = (
            {} if contract_opts is None else dict(contract_opts)
        )
        self.contract_opts.setdefault('optimize', 'auto-hq')
        if psi is not None:
            self._set_psi(psi)

    def _set_psi(self, psi):
        self._psi = psi.copy()
        if self.use_gauges:
            self.gauges0 = {}
            self._psi.gauge_all_simple_(gauges=self.gauges0)
        else:
            self.gauges0 = None

        self.tid2site = {}
        for site in self._psi.sites:
            tid, = self._psi._get_tids_from_tags(site)
            self.tid2site[tid] = site
        self.ordering = self._psi.compute_hierarchical_ordering()

    def sample(self):
        """
        """
        config = {}
        psi = self._psi.copy()

        if self.use_gauges:
            gauges = self.gauges0.copy()
        else:
            gauges = None

        omega = 1.0

        for tid in self.ordering:
            site = self.tid2site[tid]
            ind = psi.site_ind(site)

            # select a local patch
            k = psi._select_local_tids(
                [tid],
                max_distance=self.max_distance,
                fillin=0,
                virtual=False,
            )

            if self.use_gauges:
                # gauge it including dangling bonds
                k.gauge_simple_insert(gauges)

            # contract the approx reduced density matrix diagonal
            pk = (k.H & k).contract(
                ...,
                output_inds=[ind],  # directly extract diagonal
                **self.contract_opts,
            ).data

            # normalize and sample a state for this site
            pk /= pk.sum()
            idx = self.rng.choice(np.arange(2), p=pk)
            config[site] = idx

            # track the probability chain
            omega *= pk[idx]

            # fix the site to measurement
            psi.tensor_map[tid].isel_({ind: idx})

            if self.use_gauges:
                # update local gauges to take measurement into account
                psi._gauge_local_tids(
                    [tid],
                    max_distance=(self.max_distance + 1),
                    method='simple', gauges=gauges
                )

        return config, omega

    candidate = sample

    def accept(self, config):
        pass

    def update(self, **kwargs):
        self._set_psi(kwargs['psi'])


class ExchangeSampler:

    def __init__(self, edges, seed=None):
        self.edges = tuple(sorted(edges))
        self.Ne = len(self.edges)
        self.sites = sorted(set(site for edge in edges for site in edge))
        self.N = len(self.sites)
        self.rng = np.random.default_rng(seed)
        values0 = [0] * (self.N // 2) + [1] * (self.N // 2)
        if self.N % 2 == 1:
            values0.append(0)
        values0 = self.rng.permutation(values0)
        self.config = dict(zip(self.sites, values0))

    def candidate(self):
        nconfig = self.config.copy()
        for i in self.rng.permutation(np.arange(self.Ne)):
            cooa, coob = self.edges[i]
            xa, xb = nconfig[cooa], nconfig[coob]
            if xa == xb:
                continue
            nconfig[cooa], nconfig[coob] = xb, xa
            return nconfig, 1.0

    def accept(self, config):
        self.config = config

    def sample(self):
        config, omega = self.candidate()
        self.accept(config)
        return config, omega

    def update(self, **_):
        pass


class HamiltonianSampler:

    def __init__(self, ham, seed=None):
        self.ham = ham
        self.rng = np.random.default_rng(seed)

        self.N = len(self.ham.sites)
        values0 = [0] * (self.N // 2) + [1] * (self.N // 2)
        if self.N % 2 == 1:
            values0.append(0)
        values0 = self.rng.permutation(values0)
        self.config = dict(zip(self.ham.sites, values0))

    def candidate(self):
        generate = True
        while generate:
            # XXX: could do this much more efficiently with a single random
            # term
            configs, _ = self.ham.config_coupling(self.config)
            i = self.rng.integers(len(configs))
            new_config = configs[i]
            generate = (new_config == self.config)
        return new_config, 1.0

    def accept(self, config):
        self.config = config

    def sample(self):
        config, omega = self.candidate()
        self.accept(config)
        return config, omega

    def update(self, **_):
        pass


class MetropolisHastingsSampler:
    """
    """

    def __init__(
        self,
        sub_sampler,
        amplitude_factory=None,
        initial=None,
        burn_in=0,
        seed=None,
        track=False,
    ):
        self.sub_sampler = sub_sampler

        if amplitude_factory is not None:
            self.prob_fn = amplitude_factory.prob
        else:
            # will initialize later
            self.prob_fn = None

        if initial is not None:
            self.config, self.omega, self.prob = initial
        else:
            self.config = self.omega = self.prob = None

        self.seed = seed
        self.rng = np.random.default_rng(self.seed)
        self.accepted = 0
        self.total = 0
        self.burn_in = burn_in

        # should we record the history?
        self.track = track
        if self.track:
            self.omegas = array.array('d')
            self.probs = array.array('d')
            self.acceptances = array.array('d')
        else:
            self.omegas = self.probs = self.acceptances = None

    @property
    def acceptance_ratio(self):
        if self.total == 0:
            return 0.0
        return self.accepted / self.total

    def sample(self):
        if self.config is None:
            # check if we are starting from scratch
            self.config, self.omega = self.sub_sampler.sample()
            self.prob = self.prob_fn(self.config)

        while True:
            self.total += 1

            # generate candidate configuration
            nconfig, nomega = self.sub_sampler.candidate()
            nprob = self.prob_fn(nconfig)

            # compute acceptance probability
            acceptance = (nprob * self.omega) / (self.prob * nomega)

            if self.track:
                self.omegas.append(nomega)
                self.probs.append(nprob)
                self.acceptances.append(acceptance)

            if (self.rng.uniform() < acceptance):
                self.config = nconfig
                self.omega = nomega
                self.prob = nprob
                self.accepted += 1
                self.sub_sampler.accept(nconfig)

                if (self.total > self.burn_in):
                    return self.config, self.omega

    def update(self, **kwargs):
        self.prob_fn = kwargs['amplitude_factory'].prob
        self.sub_sampler.update(**kwargs)

    def plot(self):
        from matplotlib import pyplot as plt

        fig, axs = plt.subplots(ncols=2, figsize=(8, 4))
        fig.suptitle(f'acceptance ratio = {100 * self.acceptance_ratio:.2f} %')

        mins = min(self.omegas)
        maxs = max(self.omegas)

        axs[0].plot([mins, maxs], [mins, maxs], color='red')
        axs[0].scatter(self.probs, self.omegas, marker='.', alpha=0.5)
        axs[0].set_xlabel('$\pi(x)$')
        axs[0].set_ylabel('$\omega(x)$')
        axs[0].set_xscale('log')
        axs[0].set_yscale('log')
        axs[0].grid(True, c=(0.97 ,0.97, 0.97), which='major')
        axs[0].set_axisbelow(True)

        minh = np.log10(min(self.acceptances))
        maxh = np.log10(max(self.acceptances))
        axs[1].hist(self.acceptances, bins=np.logspace(minh, maxh), color='green');
        axs[1].set_xlabel('$A = \dfrac{\pi(x)\omega(y)}{\pi(y)\omega(x)}$')
        axs[1].axvline(1.0, color='orange')
        axs[1].set_xscale('log')
        axs[1].grid(True, c=(0.97 ,0.97, 0.97), which='major')
        axs[1].set_axisbelow(True)


# --------------------------------------------------------------------------- #

def auto_share_multicall(func, arrays, configs):
    """Call the function ``func``, which should be an array
    function making use of autoray dispatched calls, multiple
    times, automatically reusing shared intermediates.
    """
    with ar.lazy.shared_intermediates():
        lzarrays_all = [
            # different variants provided as first dimension
            ar.lazy.array(x) if hasattr(x, 'shape') else
            # different variants provided as a sequence
            list(map(ar.lazy.array, x))
            for x in arrays
        ]
        lzarrays_config = lzarrays_all.copy()

        outs = []
        for config in configs:
            # for each config, insert the correct inputs
            for k, v in config.items():
                lzarrays_config[k] = lzarrays_all[k][v]
            # evaluate the function
            outs.append(func(lzarrays_config))

        # combine into single output object
        final = ar.lazy.stack(tuple(outs))

    # evaluate all configs simulteneously
    return final.compute()


class ComposePartial:

    __slots__ = (
        "f",
        "f_args",
        "f_kwargs",
        "g",
    )

    def __init__(self, f, f_args, f_kwargs, g):
        self.f = f
        self.f_args = f_args
        self.f_kwargs = f_kwargs
        self.g = g

    def __call__(self, *args, **kwargs):
        y = self.g(*args, **kwargs)
        f_args = (
            y if isinstance(v, ar.lazy.LazyArray) else v
            for v in self.f_args
        )
        return self.f(*f_args, **self.f_kwargs)


_partial_compose_cache = {}


def get_compose_partial(f, f_args, f_kwargs, g):

    key = (
        f,
        tuple(
            '__placeholder__' if isinstance(v, ar.lazy.LazyArray)
            else v
            for v in f_args
        ),
        tuple(sorted(f_kwargs.items())),
        g,
    )

    try:
        fg = _partial_compose_cache[key]
    except KeyError:
        fg = _partial_compose_cache[key] = ComposePartial(f, f_args, f_kwargs, g)
    except TypeError:
        fg = ComposePartial(f, f_args, f_kwargs, g)

    return fg


def fuse_unary_ops_(Z):
    queue = [Z]
    seen = set()
    while queue:
        node = queue.pop()
        if (
            len(node._deps) == 1 and
            any(isinstance(v, ar.lazy.LazyArray) for v in node.args)
        ):
            dep, = node._deps
            if dep._nchild == 1 and dep._fn:
                node._fn = get_compose_partial(node._fn, node._args, node._kwargs, dep._fn)
                node._args = dep._args
                node._kwargs = dep._kwargs
                node._deps = dep._deps
                queue.append(node)
                continue

        for dep in node._deps:
            if dep not in seen:
                queue.append(dep)
                seen.add(dep)



class AmplitudeFactory:

    def __init__(
        self,
        psi=None,
        contract_fn=None,
        maxsize=2**20,
        autojit_opts=(),
        **contract_opts,
    ):
        from quimb.utils import LRU

        self.contract_fn = contract_fn
        self.contract_opts = contract_opts
        if self.contract_opts.get('max_bond', None) is not None:
            self.contract_opts.setdefault('cutoff', 0.0)

        self.autojit_opts = dict(autojit_opts)

        if psi is not None:
            self._set_psi(psi)

        self.store = LRU(maxsize=maxsize)
        self.hits = 0
        self.queries = 0

    def _set_psi(self, psi):
        psi0 = psi.copy()

        self.arrays = []
        self.sitemap = {}
        variables = []

        for site in psi0.sites:
            ix = psi0.site_ind(site)
            t, = psi0._inds_get(ix)

            # want variable index first
            t.moveindex_(ix, 0)
            self.sitemap[site] = len(self.arrays)
            self.arrays.append(t.data)

            # insert lazy variable for sliced tensor
            variable = ar.lazy.Variable(t.shape[1:], backend='autoray.lazy')
            variables.append(variable)
            t.modify(data=variable, inds=t.inds[1:])

        # trace the function lazily
        if self.contract_fn is None:
            Z = psi0.contract(..., output_inds=(), **self.contract_opts)
        else:
            Z = self.contract_fn(psi0, **self.contract_opts)

        # get the functional form of this traced contraction
        self.f_lazy = Z.get_function(variables)

        # this can then itself be traced with concrete arrays
        self.f = ar.autojit(self.f_lazy, **self.autojit_opts)

    def compute_single(self, config):
        """Compute the amplitude of ``config``, making use of autojit.
        """
        arrays = self.arrays.copy()
        for site, v in config.items():
            i = self.sitemap[site]
            arrays[i] = self.arrays[i][v]
        return self.f(arrays)

    def compute_multi(self, configs):
        """Compute the amplitudes corresponding to the sequence ``configs``,
        making use of shared intermediates.
        """
        # translate index config to position configs
        iconfigs = [
            {self.sitemap[site]: v for site, v in config.items()}
            for config in configs
        ]
        return auto_share_multicall(self.f_lazy, self.arrays, iconfigs)

    # def update(self, config, coeff):
    #     """Update the amplitude cache with a new configuration.
    #     """
    #     self.store[tuple(sorted(config.items()))] = coeff

    def amplitude(self, config):
        """Get the amplitude of ``config``, either from the cache or by
        computing it.
        """
        key = tuple(sorted(config.items()))
        self.queries += 1
        if key in self.store:
            self.hits += 1
            return self.store[key]

        coeff = self.compute_single(self.psi, config)

        self.store[key] = coeff
        return coeff

    def amplitudes(self, configs):
        """
        """
        # first parse out the configurations we need to compute
        all_keys = []
        new_keys = []
        new_configs = []
        for config in configs:
            key = tuple(sorted(config.items()))
            all_keys.append(key)
            self.queries += 1
            if key in self.store:
                self.hits += 1
            else:
                new_keys.append(key)
                new_configs.append(config)

        # compute the new configurations
        if new_configs:
            new_coeffs = self.compute_multi(new_configs)
            for key, coeff in zip(new_keys, new_coeffs):
                self.store[key] = coeff

        # return the full set of old and new coefficients
        return [self.store[key] for key in all_keys]

    def prob(self, config):
        """Calculate the probability of a configuration.
        """
        coeff = self.amplitude(config)
        return ar.do("abs", coeff)**2

    def clear(self):
        self.store.clear()

    def __contains__(self, config):
        return tuple(sorted(config.items())) in self.store

    def __setitem__(self, config, c):
        self.store[tuple(sorted(config.items()))] = c

    def __getitem__(self, config):
        return self.amplitude(config)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}(hits={self.hits}, "
            f"queries={self.queries})>"
        )


# class AmplitudeStore:

#     def __init__(self, psi, amp_fn, maxsize=2**20):
#         from quimb.utils import LRU
#         self.psi = psi
#         self.amp_fn = amp_fn
#         self.store = LRU(maxsize=maxsize)
#         self.hits = 0
#         self.queries = 0

#     def update(self, config, coeff):
#         """Update the amplitude cache with a new configuration.
#         """
#         self.store[tuple(sorted(config.items()))] = coeff

#     def amplitude(self, config):
#         """Calculate the amplitude of a configuration.
#         """
#         self.queries += 1

#         key = tuple(sorted(config.items()))
#         if key in self.store:
#             self.hits += 1
#             return self.store[key]

#         amp = self.amp_fn(self.psi, config)

#         self.store[key] = amp
#         return amp

#     def prob(self, config):
#         """Calculate the probability of a configuration.
#         """
#         amp = self.amplitude(config)
#         return ar.do("abs", amp)**2

#     def clear(self):
#         self.store.clear()

#     def __contains__(self, config):
#         return tuple(sorted(config.items())) in self.store

#     def __setitem__(self, config, c):
#         self.store[tuple(sorted(config.items()))] = c

#     def __getitem__(self, config):
#         return self.amplitude(config)

#     def __repr__(self):
#         return (
#             f"<{self.__class__.__name__}(hits={self.hits}, "
#             f"queries={self.queries})>"
#         )


# @autojit_tn(backend='torch', check_inputs=False)
# def contract_amplitude_tn(tni, **contract_opts):
#     if 'max_bond' in contract_opts:
#         contract_opts.setdefault('cutoff', 0.0)
#     return tni.contract(..., output_inds=(), **contract_opts)


# def compute_amplitude_aj(tn, config, **contract_opts):
#     tni = tn.isel({tn.site_ind(site): v for site, v in config.items()})
#     return contract_amplitude_tn(tni, **contract_opts)


# def compute_local_energy_aj(ham, config, amp):
#     en = 0.0
#     c_configs, c_coeffs = ham.config_coupling(config)
#     cx = amp.amplitude(config)
#     for hxy, config_y in zip(c_coeffs, c_configs):
#         cy = amp.amplitude(config_y)
#         en += hxy * cy / cx
#     return en


# def compute_amp_and_gradients(psi, config, **contract_opts):
#     import torch
#     psi_t = psi.copy()
#     psi_t.apply_to_arrays(lambda x: torch.tensor(x).requires_grad_())
#     c = compute_amplitude_aj(psi_t, config, **contract_opts)
#     c.backward()
#     c = c.item()
#     return [t.data.grad.numpy() / c for t in psi_t], c


# --------------------------------------------------------------------------- #


class GradientAccumulator:

    def __init__(self):
        self._grads_logpsi = None
        self._grads_energy = None
        self._batch_energy = None
        self._num_samples = 0

    def _init_storage(self, grads):
        self._batch_energy = 0.0
        self._grads_logpsi = [np.zeros_like(g) for g in grads]
        self._grads_energy = [np.zeros_like(g) for g in grads]

    def update(self, grads_logpsi_sample, local_energy):
        if self._batch_energy is None:
            self._init_storage(grads_logpsi_sample)

        self._batch_energy += local_energy
        for g, ge, g_i in zip(
            self._grads_logpsi, self._grads_energy, grads_logpsi_sample,
        ):
            g += g_i
            ge += g_i * local_energy
        self._num_samples += 1

    def extract_grads_energy(self):
        e = self._batch_energy / self._num_samples
        grads_energy_batch = []
        for g, ge in zip(self._grads_logpsi, self._grads_energy):
            g /= self._num_samples
            ge /= self._num_samples
            grads_energy_batch.append(ge - g * e)
            # reset storage
            g.fill(0.0)
            ge.fill(0.0)
        self._batch_energy = 0.0
        self._num_samples = 0
        return grads_energy_batch


class SGD(GradientAccumulator):

    def __init__(self, learning_rate=0.01):
        self.learning_rate = learning_rate
        super().__init__()

    def transform_gradients(self):
        return [
            self.learning_rate * g
            for g in self.extract_grads_energy()
        ]


class SignDescent(GradientAccumulator):

    def __init__(self, learning_rate=0.01):
        self.learning_rate = learning_rate
        super().__init__()

    def transform_gradients(self):
        return [
            self.learning_rate * np.sign(g)
            for g in self.extract_grads_energy()
        ]


class RandomSign(GradientAccumulator):

    def __init__(self, learning_rate=0.01):
        self.learning_rate = learning_rate
        super().__init__()

    def transform_gradients(self):
        return [
            self.learning_rate * np.sign(g) * np.random.uniform(size=g.shape)
            for g in self.extract_grads_energy()
        ]


class Adam(GradientAccumulator):

    def __init__(
        self,
        learning_rate=0.01,
        beta1=0.9,
        beta2=0.999,
        eps=1e-8,
    ):
        self.learning_rate = learning_rate
        self.beta1 = beta1
        self.beta2 = beta2
        self.eps = eps
        self._num_its = 0
        self._ms = None
        self._vs = None
        super().__init__()

    def transform_gradients(self):
        # get the standard SGD gradients
        grads = self.extract_grads_energy()

        self._num_its += 1
        if self._num_its == 1:
            # first iteration, initialize storage
            self._ms = [np.zeros_like(g) for g in grads]
            self._vs = [np.zeros_like(g) for g in grads]

        deltas = []
        for i, g in enumerate(grads):
            # first  moment estimate
            m = (1 - self.beta1) * g + self.beta1 * self._ms[i]
            # second moment estimate
            v = (1 - self.beta2) * (g**2) + self.beta2 * self._vs[i]
            # bias correction
            mhat = m / (1 - self.beta1**(self._num_its))
            vhat = v / (1 - self.beta2**(self._num_its))
            deltas.append(
                self.learning_rate * mhat / (np.sqrt(vhat) + self.eps)
            )
        return deltas


from quimb.tensor.optimize import Vectorizer


class StochasticReconfigureGradients:

    def __init__(self, delta=1e-5):
        self.delta = delta
        self.vectorizer = None
        self.gs = []

    def update(self, grads_logpsi_sample, local_energy):
        if self.vectorizer is None:
            # first call, initialize storage
            self.vectorizer = Vectorizer(grads_logpsi_sample)
        self.gs.append(self.vectorizer.pack(grads_logpsi_sample).copy())
        super().update(grads_logpsi_sample, local_energy)

    def extract_grads_energy(self):
        # number of samples
        num_samples = len(self.gs)

        gs = np.stack(self.gs)
        self.gs.clear()
        # <g_i g_j>
        S = (gs.T / num_samples) @ gs
        # minus <g_i><g_j> to get S
        g = gs.sum(axis=0) / num_samples
        S -= np.outer(g, g)

        # condition by adding to diagonal
        S.flat[::S.shape[0] + 1] += self.delta

        # the uncorrected energy gradient / 'force' vector
        y = self.vectorizer.pack(super().extract_grads_energy())

        # the corrected energy gradient, which we then unvectorize
        x = np.linalg.solve(S, y)
        return self.vectorizer.unpack(x)



class SR(SGD, StochasticReconfigureGradients):

    def __init__(self, learning_rate=0.05, delta=1e-5):
        StochasticReconfigureGradients.__init__(self, delta=delta)
        SGD().__init__(self, learning_rate=learning_rate)



class SRADAM(Adam, StochasticReconfigureGradients):

    def __init__(
        self,
        learning_rate=0.01,
        beta1=0.9,
        beta2=0.999,
        eps=1e-8,
        delta=1e-5,
    ):
        StochasticReconfigureGradients.__init__(self, delta=delta)
        Adam.__init__(
            self, learning_rate=learning_rate,
            beta1=beta1, beta2=beta2, eps=eps,
        )


# --------------------------------------------------------------------------- #

class TNVMC:

    def __init__(
        self,
        psi,
        ham,
        sampler,
        conditioner='auto',
        learning_rate=1e-2,
        optimizer='adam',
        optimizer_opts=None,
        track_window_size=1000,
        **contract_opts
    ):
        from quimb.utils import ensure_dict

        self.psi = psi.copy()
        self.ham = ham
        self.sampler = sampler

        if conditioner == 'auto':

            def conditioner(psi):
                psi.equalize_norms_(1.0)

        else:
            self.conditioner = conditioner

        if self.conditioner is not None:
            # want initial arrays to be in conditioned form so that gradients
            # are approximately consistent across runs (e.g. for momentum)
            self.conditioner(self.psi)

        optimizer_opts = ensure_dict(optimizer_opts)
        self.optimizer = {
            'adam': Adam,
            'sgd': SGD,
            'sign': SignDescent,
            'signu': RandomSign,
            'sr': SR,
            'sradam': SRADAM,
        }[optimizer.lower()](learning_rate=learning_rate, **optimizer_opts)
        self.contract_opts = contract_opts

        self.amplitude_factory = AmplitudeFactory(self.psi, **contract_opts)
        self.sampler.update(psi=self.psi, amplitude_factory=self.amplitude_factory)

        # tracking information
        self.moving_stats = MovingStatistics(track_window_size)
        self.local_energies = array.array('d')
        self.energies = array.array('d')
        self.energy_errors = array.array('d')
        self.num_tensors = self.psi.num_tensors
        self.nsites = self.psi.nsites
        self._progbar = None

    def _compute_log_gradients_torch(self, config):
        import torch
        psi_t = self.psi.copy()
        psi_t.apply_to_arrays(lambda x: torch.tensor(x).requires_grad_())
        c = self.amplitude_factory(psi_t, config)
        c.backward()
        c = c.item()
        self.amplitude_factory[config] = c
        return [t.data.grad.numpy() / c for t in psi_t]

    def _compute_local_energy(self, config):
        en = 0.0
        c_configs, c_coeffs = self.ham.config_coupling(config)
        cx = self.amplitude_factory.amplitude(config)
        for hxy, config_y in zip(c_coeffs, c_configs):
            cy = self.amplitude_factory.amplitude(config_y)
            en += hxy * cy / cx
        return en / self.nsites

    def _run(self, steps, batchsize):
        for _ in range(steps):
            for _ in range(batchsize):
                config, omega = self.sampler.sample()

                # compute and track local energy
                local_energy = self._compute_local_energy(config)
                self.local_energies.append(local_energy)
                self.moving_stats.update(local_energy)
                self.energies.append(self.moving_stats.mean)
                self.energy_errors.append(self.moving_stats.err)

                # compute the sample log amplitude gradients
                grads_logpsi_sample = self._compute_log_gradients_torch(config)

                self.optimizer.update(grads_logpsi_sample, local_energy)

                if self._progbar is not None:
                    self._progbar.update()
                    self._progbar.set_description(
                        format_number_with_error(
                            self.moving_stats.mean,
                            self.moving_stats.err))

            # apply learning rate and other transforms to gradients
            deltas = self.optimizer.transform_gradients()

            # update the actual tensors
            for t, delta in zip(self.psi.tensors, deltas):
                t.modify(data=t.data - delta)

            # reset having just performed a gradient step
            if self.conditioner is not None:
                self.conditioner(self.psi)

            self.amplitude_factory.clear()
            self.sampler.update(psi=self.psi, amplitude_factory=self.amplitude_factory)

    def run(
        self,
        total=10_000,
        batchsize=100,
        progbar=True,
    ):
        steps = total // batchsize
        total = steps * batchsize

        if progbar:
            from quimb.utils import progbar as Progbar
            self._progbar = Progbar(total=total)

        try:
            self._run(steps, batchsize)
        except KeyboardInterrupt:
            pass
        finally:
            if self._progbar is not None:
                self._progbar.close()

    def measure(
        self,
        max_samples=10_000,
        rtol=1e-4,
        progbar=True,
    ):
        from xyzpy import RunningStatistics

        rs = RunningStatistics()
        energies = array.array('d')

        if progbar:
            from quimb.utils import progbar as Progbar
            pb = Progbar(total=max_samples)
        else:
            pb = None

        try:
            for _ in range(max_samples):
                config, _ = self.sampler.sample()
                local_energy = self._compute_local_energy(config)
                rs.update(local_energy)
                energies.append(local_energy)

                if pb is not None:
                    pb.update()
                    err = rs.err
                    if err != 0.0:
                        pb.set_description(format_number_with_error(rs.mean, err))

                if 0.0 < rs.rel_err < rtol:
                    break

        except KeyboardInterrupt:
            pass
        finally:
            if pb is not None:
                pb.close()

        return rs, energies

    def plot(
        self,
        figsize=(12, 6),
        yrange_quantile=(0.01, 0.99),
    ):
        from matplotlib import pyplot as plt

        x = np.arange(len(self.local_energies))
        # these are all views
        y = np.frombuffer(self.local_energies)
        ym = np.frombuffer(self.energies)
        yerr = np.frombuffer(self.energy_errors)
        yplus = ym + yerr
        yminus = ym - yerr

        fig, ax = plt.subplots(figsize=figsize)
        ax.plot(
            x, y, '.', alpha=0.5, markersize=1.0, zorder=-10,
            color=(0.1, 0.5, .7)
        )
        ax.fill_between(
            x, yminus, yplus, alpha=0.45, color=(0.6, .8, .6)
        )
        ax.plot(
            x, ym, '-', alpha=0.9, zorder=-10, linewidth=2, color=(0.6, .8, .6)
        )

        # ax.text(0, yf[0], f'{yf[0]:.4f}', c=(0.3, .6, .3), ha='left', va='top')
        # ax.text(len(yf), yf[-1], f'{yf[-1]:.4f}', c=(0.3, .6, .3), ha='left', va='top')

        # en_su = -0.719025
        # ax.axhline(en_su, linestyle='--', c=(.9, .3, .3), lw=2)
        # ax.text(x[-1], en_su, f'D={D} SU', ha='left', va='bottom', c=(.9, .3, .3))

        # ax.axhline(-0.653179604041, linestyle='--', c=(.6, .3, .3), lw=2)
        # ax.text(x[-1], -0.653179604041, 'D=3 exact', ha='left', va='bottom', c=(.6, .3, .3))

        # ax.axhline(en_ex, linestyle='--', c=(.3, .3, .3), lw=2)
        # ax.text(x[-1], en_ex, 'ED', ha='left', va='top', c=(.3, .3, .3))

        # for vl in vlines:
        #     ax.axvline(vl, zorder=0, color=(.99, .8, .0), alpha=0.5, lw=1)

        ax.set_ylim(np.quantile(y, yrange_quantile[0]),
                    np.quantile(y, yrange_quantile[1]))
        ax.set_rasterization_zorder(0)

        ax.set_xlabel('Number of local energy evaluations')
        ax.set_ylabel('Energy per site')


        return fig, ax



"""Use stochastic Lanczos quadrature to approximate spectral function sums of
any operator which has an efficient representation of action on a vector.
"""
import functools
from math import sqrt, log2, exp, inf, nan
import random
import warnings

import numpy as np
import scipy.linalg as scla
from scipy.ndimage.filters import uniform_filter1d

from ..core import ptr, prod, vdot, njit, dot, subtract_update_, divide_update_
from ..utils import int2tup, find_library, raise_cant_find_library_function
from ..gen.rand import randn, rand_rademacher, rand_phase, seed_rand
from ..linalg.mpi_launcher import get_mpi_pool

if find_library("opt_einsum") and find_library("autoray"):
    from ..tensor.tensor_core import Tensor
    from ..tensor.tensor_1d import MatrixProductOperator
    from ..tensor.tensor_approx_spectral import construct_lanczos_tridiag_MPO
else:
    reqs = "[opt_einsum,autoray]"
    Tensor = raise_cant_find_library_function(reqs)
    construct_lanczos_tridiag_MPO = raise_cant_find_library_function(reqs)


# --------------------------------------------------------------------------- #
#                  'Lazy' representation tensor contractions                  #
# --------------------------------------------------------------------------- #


def lazy_ptr_linop(psi_ab, dims, sysa, **linop_opts):
    r"""A linear operator representing action of partially tracing a bipartite
    state, then multiplying another 'unipartite' state::

          ( | )
        +-------+
        | psi_a |   ______
        +_______+  /      \
           a|      |b     |
        +-------------+   |
        |  psi_ab.H   |   |
        +_____________+   |
                          |
        +-------------+   |
        |   psi_ab    |   |
        +_____________+   |
           a|      |b     |
            |      \______/

    Parameters
    ----------
    psi_ab : ket
        State to partially trace and dot with another ket, with
        size ``prod(dims)``.
    dims : sequence of int, optional
        The sub dimensions of ``psi_ab``.
    sysa : int or sequence of int, optional
        Index(es) of the 'a' subsystem(s) to keep.
    """
    sysa = int2tup(sysa)

    Kab = Tensor(
        np.asarray(psi_ab).reshape(dims),
        inds=[
            ("kA{}" if i in sysa else "xB{}").format(i)
            for i in range(len(dims))
        ],
    )

    Bab = Tensor(
        Kab.data.conjugate(),
        inds=[
            ("bA{}" if i in sysa else "xB{}").format(i)
            for i in range(len(dims))
        ],
    )

    return (Kab & Bab).aslinearoperator(
        [f"kA{i}" for i in sysa], [f"bA{i}" for i in sysa], **linop_opts
    )


def lazy_ptr_ppt_linop(psi_abc, dims, sysa, sysb, **linop_opts):
    r"""A linear operator representing action of partially tracing a tripartite
    state, partially transposing the remaining bipartite state, then
    multiplying another bipartite state::

             ( | )
        +--------------+
        |   psi_ab     |
        +______________+  _____
         a|  ____   b|   /     \
          | /   a\   |   |c    |
          | | +-------------+  |
          | | |  psi_abc.H  |  |
          \ / +-------------+  |
           X                   |
          / \ +-------------+  |
          | | |   psi_abc   |  |
          | | +-------------+  |
          | \____/a  |b  |c    |
         a|          |   \_____/

    Parameters
    ----------
    psi_abc : ket
        State to partially trace, partially transpose, then dot with another
        ket, with size ``prod(dims)``.
        ``prod(dims[sysa] + dims[sysb])``.
    dims : sequence of int
        The sub dimensions of ``psi_abc``.
    sysa : int or sequence of int, optional
        Index(es) of the 'a' subsystem(s) to keep, with respect to all
        the dimensions, ``dims``, (i.e. pre-partial trace).
    sysa : int or sequence of int, optional
        Index(es) of the 'b' subsystem(s) to keep, with respect to all
        the dimensions, ``dims``, (i.e. pre-partial trace).
    """
    sysa, sysb = int2tup(sysa), int2tup(sysb)
    sys_ab = sorted(sysa + sysb)

    Kabc = Tensor(
        np.asarray(psi_abc).reshape(dims),
        inds=[
            ("kA{}" if i in sysa else "kB{}" if i in sysb else "xC{}").format(
                i
            )
            for i in range(len(dims))
        ],
    )

    Babc = Tensor(
        Kabc.data.conjugate(),
        inds=[
            ("bA{}" if i in sysa else "bB{}" if i in sysb else "xC{}").format(
                i
            )
            for i in range(len(dims))
        ],
    )

    return (Kabc & Babc).aslinearoperator(
        [("bA{}" if i in sysa else "kB{}").format(i) for i in sys_ab],
        [("kA{}" if i in sysa else "bB{}").format(i) for i in sys_ab],
        **linop_opts,
    )


# --------------------------------------------------------------------------- #
#                         Lanczos tri-diag technique                          #
# --------------------------------------------------------------------------- #


def inner(a, b):
    """Inner product between two vectors"""
    return vdot(a, b).real


def norm_fro(a):
    """'Frobenius' norm of a vector."""
    return sqrt(inner(a, a))


def norm_fro_approx(A, **kwargs):
    r"""Calculate the approximate frobenius norm of any hermitian linear
    operator:

    .. math::

        \mathrm{Tr} \left[ A^{\dagger} A \right]

    Parameters
    ----------
    A : linear operator like
        Operator with a dot method, assumed to be hermitian, to estimate the
        frobenius norm of.
    kwargs
        Supplied to :func:`approx_spectral_function`.

    Returns
    -------
    float
    """
    return approx_spectral_function(A, lambda x: x**2, **kwargs) ** 0.5


def random_rect(
    shape,
    dist="rademacher",
    orthog=False,
    norm=True,
    seed=False,
    dtype=complex,
):
    """Generate a random array optionally orthogonal.

    Parameters
    ----------
    shape : tuple of int
        The shape of array.
    dist : {'guassian', 'rademacher'}
        Distribution of the random variables.
    orthog : bool or operator.
        Orthogonalize the columns if more than one.
    norm : bool
        Explicitly normalize the frobenius norm to 1.
    """
    if seed:
        # needs to be truly random so e.g. MPI processes don't overlap
        seed_rand(random.SystemRandom().randint(0, 2**32 - 1))

    if dist == "rademacher":
        V = rand_rademacher(shape, scale=1 / sqrt(prod(shape)), dtype=dtype)
        # already normalized

    elif dist == "gaussian":
        V = randn(
            shape, scale=1 / (prod(shape) ** 0.5 * 2**0.5), dtype=dtype
        )
        if norm:
            V /= norm_fro(V)

    elif dist == "phase":
        V = rand_phase(shape, scale=1 / sqrt(prod(shape)), dtype=dtype)
        # already normalized

    else:
        raise ValueError(f"`dist={dist}` not understood.")

    if orthog and min(shape) > 1:
        V = scla.orth(V)
        V /= sqrt(min(V.shape))

    return V


def construct_lanczos_tridiag(
    A,
    K,
    v0=None,
    bsz=1,
    k_min=10,
    orthog=False,
    beta_tol=1e-6,
    seed=False,
    v0_opts=None,
):
    """Construct the tridiagonal lanczos matrix using only matvec operators.
    This is a generator that iteratively yields the alpha and beta digaonals
    at each step.

    Parameters
    ----------
    A : dense array, sparse matrix or linear operator
        The operator to approximate, must implement ``.dot`` method to compute
        its action on a vector.
    K : int, optional
        The maximum number of iterations and thus rank of the matrix to find.
    v0 : vector, optional
        The starting vector to iterate with, default to random.
    bsz : int, optional
        The block size (number of columns) of random vectors to iterate with.
    k_min : int, optional
        The minimum size of the krylov subspace for form.
    orthog : bool, optional
        If True, perform full re-orthogonalization for each new vector.
    beta_tol : float, optional
        The 'breakdown' tolerance. If the next beta ceofficient in the lanczos
        matrix is less that this, implying that the full non-null space has
        been found, terminate early.
    seed : bool, optional
        If True, seed the numpy random generator with a system random int.

    Yields
    ------
    alpha : sequence of float of length k
        The diagonal entries of the lanczos matrix.
    beta : sequence of float of length k
        The off-diagonal entries of the lanczos matrix, with the last entry
        the 'look' forward value.
    scaling : float
        How to scale the overall weights.
    """
    d = A.shape[0]

    if bsz == 1:
        v_shp = (d,)
    else:
        orthog = False
        v_shp = (d, bsz)

    alpha = np.zeros(K + 1, dtype=get_equivalent_real_dtype(A.dtype))
    beta = np.zeros(K + 2, dtype=get_equivalent_real_dtype(A.dtype))
    beta[1] = sqrt(prod(v_shp))  # by construction

    if v0 is None:
        if v0_opts is None:
            v0_opts = {}
        q = random_rect(v_shp, seed=seed, dtype=A.dtype, **v0_opts)
    else:
        q = v0.astype(A.dtype)
        divide_update_(q, norm_fro(q), q)
    v = np.zeros_like(q)

    if orthog:
        Q = np.copy(q).reshape(-1, 1)

    for j in range(1, K + 1):

        r = dot(A, q)
        subtract_update_(r, beta[j], v)
        alpha[j] = inner(q, r)
        subtract_update_(r, alpha[j], q)

        # perform full orthogonalization
        if orthog:
            r -= Q.dot(Q.conj().T.dot(r))

        beta[j + 1] = norm_fro(r)

        # check for convergence
        if abs(beta[j + 1]) < beta_tol:
            yield (
                alpha[1 : j + 1].copy(),
                beta[2 : j + 2].copy(),
                beta[1] ** 2 / bsz,
            )
            break

        v[()] = q
        divide_update_(r, beta[j + 1], q)

        # keep all vectors
        if orthog:
            Q = np.concatenate((Q, q.reshape(-1, 1)), axis=1)

        if j >= k_min:
            yield (
                alpha[1 : j + 1].copy(),
                beta[2 : j + 2].copy(),
                beta[1] ** 2 / bsz,
            )


def lanczos_tridiag_eig(alpha, beta, check_finite=True):
    """Find the eigen-values and -vectors of the Lanczos triadiagonal matrix.

    Parameters
    ----------
    alpha : array of float
        The diagonal.
    beta : array of float
        The k={-1, 1} off-diagonal. Only first ``len(alpha) - 1`` entries used.
    """
    Tk_banded = np.empty((2, alpha.size), dtype=alpha.dtype)
    Tk_banded[1, -1] = 0.0  # sometimes can get nan here? -> breaks eig_banded
    Tk_banded[0, :] = alpha
    Tk_banded[1, : beta.size] = beta

    try:
        tl, tv = scla.eig_banded(
            Tk_banded, lower=True, check_finite=check_finite
        )

    # sometimes get no convergence -> use dense hermitian method
    except scla.LinAlgError:  # pragma: no cover
        tl, tv = np.linalg.eigh(
            np.diag(alpha) + np.diag(beta[: alpha.size - 1], -1), UPLO="L"
        )

    return tl, tv


def calc_trace_fn_tridiag(tl, tv, f, pos=True):
    """Spectral ritz function sum, weighted by ritz vectors."""
    return sum(
        tv[0, i] ** 2 * f(max(tl[i], 0.0) if pos else tl[i])
        for i in range(tl.size)
    )


@njit
def ext_per_trim(x, p=0.6, s=1.0):  # pragma: no cover
    r"""Extended percentile trimmed-mean. Makes the mean robust to asymmetric
    outliers, while using all data when it is nicely clustered. This can be
    visualized roughly as::

            |--------|=========|--------|
        x     x   xx xx xxxxx xxx   xx     x      x           x

    Where the inner range contains the central ``p`` proportion of the data,
    and the outer ranges entends this by a factor of ``s`` either side.

    Parameters
    ----------
    x : array
        Data to trim.
    p : Proportion of data used to define the 'central' percentile.
        For example, p=0.5 gives the inter-quartile range.
    s : Include data up to this factor times the central 'percentile' range
        away from the central percentile itself.

    Returns
    xt : array
        Trimmed data.
    """
    lb = np.percentile(x, 100 * (1 - p) / 2)
    ub = np.percentile(x, 100 * (1 + p) / 2)
    ib = ub - lb

    trimmed_x = x[(lb - s * ib < x) & (x < ub + s * ib)]

    return trimmed_x


@njit  # pragma: no cover
def nbsum(xs):
    tot = 0
    for x in xs:
        tot += x
    return tot


@njit  # pragma: no cover
def std(xs):
    """Simple standard deviation - don't invoke numpy for small lists."""
    N = len(xs)
    xm = nbsum(xs) / N
    var = nbsum([(x - xm) ** 2 for x in xs]) / N
    return var**0.5


def calc_est_fit(estimates, conv_n, tau):
    """Make estimate by fitting exponential convergence to estimates."""
    n = len(estimates)

    if n < conv_n:
        return nan, inf

    # iteration number, fit function to inverse this to get k->infinity
    ks = np.arange(1, len(estimates) + 1)

    # smooth data with a running mean
    smoothed_estimates = uniform_filter1d(estimates, n // 2)

    # ignore this amount of the initial estimates and fit later part only
    ni = n // 2

    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            # fit the inverse data with a line, weighting recent ests more
            popt, pcov = np.polyfit(
                x=(1 / ks[ni:]),
                y=smoothed_estimates[ni:],
                w=ks[ni:],
                deg=1,
                cov=True,
            )

        # estimate of function at 1 / k = 0 and standard error
        est, err = popt[-1], abs(pcov[-1, -1]) ** 0.5

    except (ValueError, RuntimeError):
        est, err = nan, inf

    return est, err


def calc_est_window(estimates, mean_ests, conv_n):
    """Make estimate from mean of last ``m`` samples, following:

    1. Take between ``conv_n`` and 12 estimates.
    2. Pair the estimates as they are alternate upper/lower bounds
    3. Compute the standard error on the paired estimates.
    """
    m_est = min(max(conv_n, len(estimates) // 8), 12)

    est = sum(estimates[-m_est:]) / len(estimates[-m_est:])
    mean_ests.append(est)

    if len(estimates) > conv_n:
        # check for convergence using variance of paired last m estimates
        #   -> paired because estimates alternate between upper and lower bound
        paired_ests = tuple(
            (a + b) / 2
            for a, b in zip(estimates[-m_est::2], estimates[-m_est + 1 :: 2])
        )
        err = std(paired_ests) / (m_est / 2) ** 0.5
    else:
        err = inf

    return est, err


def single_random_estimate(
    A,
    K,
    bsz,
    beta_tol,
    v0,
    f,
    pos,
    tau,
    tol_scale,
    k_min=10,
    verbosity=0,
    *,
    seed=None,
    v0_opts=None,
    **lanczos_opts,
):
    # choose normal (any LinearOperator) or MPO lanczos tridiag construction
    if isinstance(A, MatrixProductOperator):
        lanc_fn = construct_lanczos_tridiag_MPO
    else:
        lanc_fn = construct_lanczos_tridiag
        lanczos_opts["bsz"] = bsz

    estimates = []
    mean_ests = []

    # the number of samples to check standard deviation convergence with
    conv_n = 6  # 3 pairs

    # iteratively build the lanczos matrix, checking for convergence
    for alpha, beta, scaling in lanc_fn(
        A,
        K=K,
        beta_tol=beta_tol,
        seed=seed,
        k_min=k_min - 2 * conv_n,
        v0=v0() if callable(v0) else v0,
        v0_opts=v0_opts,
        **lanczos_opts,
    ):

        try:
            Tl, Tv = lanczos_tridiag_eig(alpha, beta, check_finite=False)
            Gf = scaling * calc_trace_fn_tridiag(Tl, Tv, f=f, pos=pos)
        except scla.LinAlgError:  # pragma: no cover
            warnings.warn("Approx Spectral Gf tri-eig didn't converge.")
            estimates.append(np.nan)
            continue

        k = alpha.size
        estimates.append(Gf)

        # check for break-down convergence (e.g. found entire subspace)
        #     in which case latest estimate should be accurate
        if abs(beta[-1]) < beta_tol:
            if verbosity >= 2:
                print(f"k={k}: Beta breadown, returning {Gf}.")
            est = Gf
            break

        # compute an estimate and error using a window of the last few results
        win_est, win_err = calc_est_window(estimates, mean_ests, conv_n)

        # try and compute an estimate and error using exponential fit
        fit_est, fit_err = calc_est_fit(mean_ests, conv_n, tau)

        # take whichever has lowest error
        est, err = min(
            (win_est, win_err),
            (fit_est, fit_err),
            key=lambda est_err: est_err[1],
        )
        converged = err < tau * (abs(win_est) + tol_scale)

        if verbosity >= 2:
            if verbosity >= 3:
                print(f"est_win={win_est}, err_win={win_err}")
                print(f"est_fit={fit_est}, err_fit={fit_err}")

            print(f"k={k}: Gf={Gf}, Est={est}, Err={err}")
            if converged:
                print(f"k={k}: Converged to tau {tau}.")

        if converged:
            break

    if verbosity >= 1:
        print(f"k={k}: Returning estimate {est}.")

    return est


def calc_stats(samples, mean_p, mean_s, tol, tol_scale):
    """Get an estimate from samples."""
    samples = np.array(samples)

    xtrim = ext_per_trim(samples, p=mean_p, s=mean_s)

    # sometimes everything is an outlier...
    if xtrim.size == 0:  # pragma: no cover
        estimate, sdev = np.mean(samples), std(samples)
    else:
        estimate, sdev = np.mean(xtrim), std(xtrim)

    err = sdev / len(samples) ** 0.5

    converged = err < tol * (abs(estimate) + tol_scale)

    return estimate, err, converged


def get_single_precision_dtype(dtype):
    if np.issubdtype(dtype, np.complexfloating):
        return np.complex64
    elif np.issubdtype(dtype, np.floating):
        return np.float32
    else:
        raise ValueError(f"dtype {dtype} not understood.")


def get_equivalent_real_dtype(dtype):
    if dtype in ("float64", "complex128"):
        return "float64"
    elif dtype in ("float32", "complex64"):
        return "float32"
    else:
        raise ValueError(f"dtype {dtype} not understood.")


def approx_spectral_function(
    A,
    f,
    tol=1e-2,
    *,
    bsz=1,
    R=1024,
    tol_scale=1,
    tau=1e-4,
    k_min=10,
    k_max=512,
    beta_tol=1e-6,
    mpi=False,
    mean_p=0.7,
    mean_s=1.0,
    pos=False,
    v0=None,
    verbosity=0,
    single_precision="AUTO",
    info=None,
    **lanczos_opts,
):
    """Approximate a spectral function, that is, the quantity ``Tr(f(A))``.

    Parameters
    ----------
    A : dense array, sparse matrix or LinearOperator
        Operator to approximate spectral function for. Should implement
        ``A.dot(vec)``.
    f : callable
        Scalar function with which to act on approximate eigenvalues.
    tol : float, optional
        Relative convergence tolerance threshold for error on mean of repeats.
        This can pretty much be relied on as the overall accuracy. See also
        ``tol_scale`` and ``tau``. Default: 1%.
    bsz : int, optional
        Number of simultenous vector columns to use at once, 1 equating to the
        standard lanczos method. If ``bsz > 1`` then ``A`` must implement
        matrix-matrix multiplication. This is a more performant way of
        essentially increasing ``R``, at the cost of more memory. Default: 1.
    R : int, optional
        The number of repeats with different initial random vectors to perform.
        Increasing this should increase accuracy as ``sqrt(R)``. Cost of
        algorithm thus scales linearly with ``R``. If ``tol`` is non-zero, this
        is the maximum number of repeats.
    tau : float, optional
        The relative tolerance required for a single lanczos run to converge.
        This needs to be small enough that each estimate with a single random
        vector produces an unbiased sample of the operators spectrum..
    k_min : int, optional
        The minimum size of the krylov subspace to form for each sample.
    k_max : int, optional
        The maximum size of the kyrlov space to form. Cost of algorithm scales
        linearly with ``K``. If ``tau`` is non-zero, this is the maximum size
        matrix to form.
    tol_scale : float, optional
        This sets the overall expected scale of each estimate, so that an
        absolute tolerance can be used for values near zero. Default: 1.
    beta_tol : float, optional
        The 'breakdown' tolerance. If the next beta ceofficient in the lanczos
        matrix is less that this, implying that the full non-null space has
        been found, terminate early. Default: 1e-6.
    mpi : bool, optional
        Whether to parallelize repeat runs over MPI processes.
    mean_p : float, optional
        Factor for robustly finding mean and err of repeat estimates,
        see :func:`ext_per_trim`.
    mean_s : float, optional
        Factor for robustly finding mean and err of repeat estimates,
        see :func:`ext_per_trim`.
    v0 : vector, or callable
        Initial vector to iterate with, sets ``R=1`` if given. If callable, the
        function to produce a random intial vector (sequence).
    pos : bool, optional
        If True, make sure any approximate eigenvalues are positive by
        clipping below 0.
    verbosity : {0, 1, 2}, optional
        How much information to print while computing.
    single_precision : {'AUTO', False, True}, optional
        Try and convert the operator to single precision. This can lead to much
        faster operation, especially if a GPU is available. Additionally,
        double precision is not really needed given the stochastic nature of
        the algorithm.
    lanczos_opts
        Supplied to
        :func:`~quimb.linalg.approx_spectral.single_random_estimate` or
        :func:`~quimb.linalg.approx_spectral.construct_lanczos_tridiag`.


    Returns
    -------
    scalar
        The approximate value ``Tr(f(a))``.

    See Also
    --------
    construct_lanczos_tridiag
    """
    if single_precision == "AUTO":
        single_precision = hasattr(A, "astype")
    if single_precision:
        A = A.astype(get_single_precision_dtype(A.dtype))

    if (v0 is not None) and not callable(v0):
        # we only have one sample to run
        R = 1
    else:
        R = max(1, int(R / bsz))

    if tau is None:
        # require better precision for the lanczos procedure, otherwise biased
        tau = tol / 1000

    if verbosity:
        print(f"LANCZOS f(A) CALC: tol={tol}, tau={tau}, R={R}, bsz={bsz}")

    # generate repeat estimates
    kwargs = {
        "A": A,
        "K": k_max,
        "bsz": bsz,
        "beta_tol": beta_tol,
        "v0": v0,
        "f": f,
        "pos": pos,
        "tau": tau,
        "k_min": k_min,
        "tol_scale": tol_scale,
        "verbosity": verbosity,
        **lanczos_opts,
    }

    if not mpi:

        def gen_results():
            for _ in range(R):
                yield single_random_estimate(**kwargs)

    else:
        pool = get_mpi_pool()
        kwargs["seed"] = True
        fs = [pool.submit(single_random_estimate, **kwargs) for _ in range(R)]

        def gen_results():
            for f in fs:
                yield f.result()

    # iterate through estimates, waiting for convergence
    results = gen_results()
    estimate = None
    samples = []
    for _ in range(R):
        samples.append(next(results))

        if verbosity >= 1:
            print(f"Repeat {len(samples)}: estimate is {samples[-1]}")

        # wait a few iterations before checking error on mean breakout
        if len(samples) >= 3:
            estimate, err, converged = calc_stats(
                samples, mean_p, mean_s, tol, tol_scale
            )
            if verbosity >= 1:
                print(f"Total estimate = {estimate} ± {err}")
            if converged:
                if verbosity >= 1:
                    print(f"Repeat {len(samples)}: converged to tol {tol}")
                break

    if mpi:
        # deal with remaining futures
        extra_futures = []
        for f in fs:
            if f.done() or f.running():
                extra_futures.append(f)
            else:
                f.cancel()

        if extra_futures:
            # might as well combine finished samples
            samples.extend(f.result() for f in extra_futures)
            estimate, err, converged = calc_stats(
                samples, mean_p, mean_s, tol, tol_scale
            )

    if estimate is None:
        estimate, err, _ = calc_stats(samples, mean_p, mean_s, tol, tol_scale)

    if verbosity >= 1:
        print(f"ESTIMATE is {estimate} ± {err}")

    if info is not None:
        if "samples" in info:
            info["samples"] = samples
        if "error" in info:
            info["error"] = err

    return estimate


@functools.wraps(approx_spectral_function)
def tr_abs_approx(*args, **kwargs):
    return approx_spectral_function(*args, f=abs, **kwargs)


@functools.wraps(approx_spectral_function)
def tr_exp_approx(*args, **kwargs):
    return approx_spectral_function(*args, f=exp, **kwargs)


@functools.wraps(approx_spectral_function)
def tr_sqrt_approx(*args, **kwargs):
    return approx_spectral_function(*args, f=sqrt, pos=True, **kwargs)


def xlogx(x):
    return x * log2(x) if x > 0 else 0.0


@functools.wraps(approx_spectral_function)
def tr_xlogx_approx(*args, **kwargs):
    return approx_spectral_function(*args, f=xlogx, **kwargs)


# --------------------------------------------------------------------------- #
#                             Specific quantities                             #
# --------------------------------------------------------------------------- #


def entropy_subsys_approx(psi_ab, dims, sysa, backend=None, **kwargs):
    """Approximate the (Von Neumann) entropy of a pure state's subsystem.

    Parameters
    ----------
    psi_ab : ket
        Bipartite state to partially trace and find entopy of.
    dims : sequence of int, optional
        The sub dimensions of ``psi_ab``.
    sysa : int or sequence of int, optional
        Index(es) of the 'a' subsystem(s) to keep.
    kwargs
        Supplied to :func:`approx_spectral_function`.
    """
    lo = lazy_ptr_linop(psi_ab, dims=dims, sysa=sysa, backend=backend)
    return -tr_xlogx_approx(lo, **kwargs)


def tr_sqrt_subsys_approx(psi_ab, dims, sysa, backend=None, **kwargs):
    """Approximate the trace sqrt of a pure state's subsystem.

    Parameters
    ----------
    psi_ab : ket
        Bipartite state to partially trace and find trace sqrt of.
    dims : sequence of int, optional
        The sub dimensions of ``psi_ab``.
    sysa : int or sequence of int, optional
        Index(es) of the 'a' subsystem(s) to keep.
    kwargs
        Supplied to :func:`approx_spectral_function`.
    """
    lo = lazy_ptr_linop(psi_ab, dims=dims, sysa=sysa, backend=backend)
    return tr_sqrt_approx(lo, **kwargs)


def norm_ppt_subsys_approx(psi_abc, dims, sysa, sysb, backend=None, **kwargs):
    """Estimate the norm of the partial transpose of a pure state's subsystem."""
    lo = lazy_ptr_ppt_linop(
        psi_abc, dims=dims, sysa=sysa, sysb=sysb, backend=backend
    )
    return tr_abs_approx(lo, **kwargs)


def logneg_subsys_approx(psi_abc, dims, sysa, sysb, **kwargs):
    """Estimate the logarithmic negativity of a pure state's subsystem.

    Parameters
    ----------
    psi_abc : ket
        Pure tripartite state, for which estimate the entanglement between
        'a' and 'b'.
    dims : sequence of int
        The sub dimensions of ``psi_abc``.
    sysa : int or sequence of int, optional
        Index(es) of the 'a' subsystem(s) to keep, with respect to all
        the dimensions, ``dims``, (i.e. pre-partial trace).
    sysa : int or sequence of int, optional
        Index(es) of the 'b' subsystem(s) to keep, with respect to all
        the dimensions, ``dims``, (i.e. pre-partial trace).
    kwargs
        Supplied to :func:`approx_spectral_function`.
    """
    nrm = norm_ppt_subsys_approx(psi_abc, dims, sysa, sysb, **kwargs)
    return max(0.0, log2(nrm))


def negativity_subsys_approx(psi_abc, dims, sysa, sysb, **kwargs):
    """Estimate the negativity of a pure state's subsystem.

    Parameters
    ----------
    psi_abc : ket
        Pure tripartite state, for which estimate the entanglement between
        'a' and 'b'.
    dims : sequence of int
        The sub dimensions of ``psi_abc``.
    sysa : int or sequence of int, optional
        Index(es) of the 'a' subsystem(s) to keep, with respect to all
        the dimensions, ``dims``, (i.e. pre-partial trace).
    sysa : int or sequence of int, optional
        Index(es) of the 'b' subsystem(s) to keep, with respect to all
        the dimensions, ``dims``, (i.e. pre-partial trace).
    kwargs
        Supplied to :func:`approx_spectral_function`.
    """
    nrm = norm_ppt_subsys_approx(psi_abc, dims, sysa, sysb, **kwargs)
    return max(0.0, (nrm - 1) / 2)


def gen_bipartite_spectral_fn(exact_fn, approx_fn, pure_default):
    """Generate a function that computes a spectral quantity of the subsystem
    of a pure state. Automatically computes for the smaller subsystem, or
    switches to the approximate method for large subsystems.

    Parameters
    ----------
    exact_fn : callable
        The function that computes the quantity on a density matrix, with
        signature: ``exact_fn(rho_a, rank=...)``.
    approx_fn : callable
        The function that approximately computes the quantity using a lazy
        representation of the whole system. With signature
        ``approx_fn(psi_ab, dims, sysa, **approx_opts)``.
    pure_default : float
        The default value when the whole state is the subsystem.

    Returns
    -------
    bipartite_spectral_fn : callable
        The function, with signature:
        ``(psi_ab, dims, sysa, approx_thresh=2**13, **approx_opts)``
    """

    def bipartite_spectral_fn(
        psi_ab, dims, sysa, approx_thresh=2**13, **approx_opts
    ):
        sysa = int2tup(sysa)
        sz_a = prod(d for i, d in enumerate(dims) if i in sysa)
        sz_b = prod(dims) // sz_a

        # pure state
        if sz_b == 1:
            return pure_default

        # also check if system b is smaller, since spectrum is same for both
        if sz_b < sz_a:
            # if so swap things around
            sz_a = sz_b
            sysb = [i for i in range(len(dims)) if i not in sysa]
            sysa = sysb

        # check whether to use approx lanczos method
        if (approx_thresh is not None) and (sz_a >= approx_thresh):
            return approx_fn(psi_ab, dims, sysa, **approx_opts)

        rho_a = ptr(psi_ab, dims, sysa)
        return exact_fn(rho_a)

    return bipartite_spectral_fn



from .tensor_3d import gen_3d_bonds
from .tensor_arbgeom_tebd import LocalHamGen


class LocalHam3D(LocalHamGen):

    def __init__(self, Lx, Ly, Lz, H2, H1=None):
        self.Lx = int(Lx)
        self.Ly = int(Ly)
        self.Lz = int(Lz)

        # parse two site terms
        if hasattr(H2, 'shape'):
            # use as default nearest neighbour term
            H2 = {None: H2}
        else:
            H2 = dict(H2)

        # possibly fill in default gates
        default_H2 = H2.pop(None, None)
        if default_H2 is not None:
            for coo_a, coo_b in gen_3d_bonds(Lx, Ly, Lz, steppers=[
                lambda i, j, k: (i, j, k + 1),
                lambda i, j, k: (i, j + 1, k),
                lambda i, j, k: (i + 1, j, k),
            ]):
                if (coo_a, coo_b) not in H2 and (coo_b, coo_a) not in H2:
                    H2[coo_a, coo_b] = default_H2

        super().__init__(H2=H2, H1=H1)

    @property
    def nsites(self):
        """The number of sites in the system.
        """
        return self.Lx * self.Ly * self.Lz

    def __repr__(self):
        s = "<LocalHam3D(Lx={}, Ly={}, Lx={}, num_terms={})>"
        return s.format(self.Lx, self.Ly, self.Lz, len(self.terms))



import pytest
import numpy as np
from numpy.testing import assert_allclose

from quimb import (
    rand_herm,
    rand_ket,
    eigh,
    can_use_mpi_pool,
)

from quimb.linalg import SLEPC4PY_FOUND
from quimb.linalg.scipy_linalg import eigs_scipy

if SLEPC4PY_FOUND:
    from quimb.linalg.mpi_launcher import (
        eigs_slepc_spawn,
        svds_slepc_spawn,
        mfn_multiply_slepc_spawn,
        ALREADY_RUNNING_AS_MPI,
        NUM_MPI_WORKERS,
    )

slepc4py_test = pytest.mark.skipif(
    not SLEPC4PY_FOUND, reason="No SLEPc4py installation")

mpipooltest = pytest.mark.skipif(
    not can_use_mpi_pool(), reason="Not allowed to use MPI pool.")

num_workers_to_try = [None, 1, 2, 3]


@pytest.fixture
def bigsparsemat():
    import numpy as np
    np.random.seed(42)
    return rand_herm(100, sparse=True, density=0.1)


@pytest.fixture
def big_vec():
    import numpy as np
    np.random.seed(2442)
    return rand_ket(100)


@slepc4py_test
class TestSLEPcMPI:
    @pytest.mark.parametrize("num_workers", num_workers_to_try)
    def test_eigs(self, num_workers, bigsparsemat):

        if ((num_workers is not None) and
                ALREADY_RUNNING_AS_MPI and
                num_workers > 1 and
                num_workers != NUM_MPI_WORKERS):
            with pytest.raises(ValueError):
                eigs_slepc_spawn(bigsparsemat, k=6, num_workers=num_workers)

        else:
            el, ev = eigs_slepc_spawn(bigsparsemat, k=6,
                                      num_workers=num_workers)
            elex, evex = eigs_scipy(bigsparsemat, k=6)
            assert_allclose(el, elex)
            assert_allclose(np.abs(ev.H @ evex), np.eye(6), atol=1e-7)

    @pytest.mark.parametrize("num_workers", num_workers_to_try)
    def test_expm_multiply(self, num_workers, bigsparsemat, big_vec):
        a = bigsparsemat
        k = big_vec

        if ((num_workers is not None) and
                ALREADY_RUNNING_AS_MPI and
                num_workers > 1 and
                num_workers != NUM_MPI_WORKERS):
            with pytest.raises(ValueError):
                mfn_multiply_slepc_spawn(a, k, num_workers=num_workers)

        else:
            out = mfn_multiply_slepc_spawn(a, k, num_workers=num_workers)
            al, av = eigh(a.A)
            expected = av @ np.diag(np.exp(al)) @ av.conj().T @ k
            assert_allclose(out, expected)

    @pytest.mark.parametrize("num_workers", num_workers_to_try)
    def test_svds(self, num_workers):
        a = np.random.randn(13, 7) + 1.0j * np.random.randn(13, 7)

        if ((num_workers is not None) and
                ALREADY_RUNNING_AS_MPI and
                num_workers > 1 and
                num_workers != NUM_MPI_WORKERS):
            with pytest.raises(ValueError):
                svds_slepc_spawn(a, return_vecs=True, num_workers=num_workers)

        else:
            u, s, v = svds_slepc_spawn(a, return_vecs=True,
                                       num_workers=num_workers)


@slepc4py_test
@mpipooltest
class TestMPIPool:
    def test_spawning_pool_in_pool(self, bigsparsemat):
        from quimb.linalg.mpi_launcher import get_mpi_pool
        l1 = eigs_slepc_spawn(bigsparsemat, k=6, return_vecs=False)
        pool = get_mpi_pool()
        f = pool.submit(eigs_slepc_spawn, bigsparsemat,
                        k=6, return_vecs=False, num_workers=1)
        l2 = f.result()
        assert_allclose(l1, l2)



import functools
from operator import add

from autoray import do, dag

from ..utils import check_opt, deprecated, ensure_dict
from ..utils import progbar as Progbar
from ..core import prod
from .tensor_core import (
    Tensor,
    TensorNetwork,
    get_symbol,
    rand_uuid,
    IsoTensor,
    oset,
    oset_union,
    tags_to_oset,
)
from . import array_ops as ops


def get_coordinate_formatter(ndims):
    return ",".join("{}" for _ in range(ndims))


def tensor_network_align(*tns, ind_ids=None, trace=False, inplace=False):
    r"""Align an arbitrary number of tensor networks in a stack-like geometry::

        a-a-a-a-a-a-a-a-a-a-a-a-a-a-a-a-a-a
        | | | | | | | | | | | | | | | | | | <- ind_ids[0] (defaults to 1st id)
        b-b-b-b-b-b-b-b-b-b-b-b-b-b-b-b-b-b
        | | | | | | | | | | | | | | | | | | <- ind_ids[1]
                       ...
        | | | | | | | | | | | | | | | | | | <- ind_ids[-2]
        y-y-y-y-y-y-y-y-y-y-y-y-y-y-y-y-y-y
        | | | | | | | | | | | | | | | | | | <- ind_ids[-1]
        z-z-z-z-z-z-z-z-z-z-z-z-z-z-z-z-z-z

    Parameters
    ----------
    tns : sequence of TensorNetwork
        The TNs to align, should be structured and either effective 'vectors'
        (have a ``site_ind_id``) or 'operators' (have a ``up_ind_id`` and
        ``lower_ind_id``).
    ind_ids : None, or sequence of str
        String with format specifiers to id each level of sites with. Will be
        automatically generated like ``(tns[0].site_ind_id, "__ind_a{}__",
        "__ind_b{}__", ...)`` if not given.
    inplace : bool
        Whether to modify the input tensor networks inplace.

    Returns
    -------
    tns_aligned : sequence of TensorNetwork
    """
    if not inplace:
        tns = [tn.copy() for tn in tns]

    n = len(tns)
    coordinate_formatter = get_coordinate_formatter(tns[0]._NDIMS)

    if ind_ids is None:
        if hasattr(tns[0], "site_ind_id"):
            ind_ids = [tns[0].site_ind_id]
        else:
            ind_ids = [tns[0].lower_ind_id]
        ind_ids.extend(
            f"__ind_{get_symbol(i)}{coordinate_formatter}__"
            for i in range(n - 2)
        )
    else:
        ind_ids = tuple(ind_ids)

    for i, tn in enumerate(tns):
        if hasattr(tn, "site_ind_id"):
            if i == 0:
                tn.site_ind_id = ind_ids[i]
            elif i == n - 1:
                tn.site_ind_id = ind_ids[i - 1]
            else:
                raise ValueError("An TN 'vector' can only be aligned as the "
                                 "first or last TN in a sequence.")

        elif hasattr(tn, "upper_ind_id") and hasattr(tn, "lower_ind_id"):
            if i != 0:
                tn.upper_ind_id = ind_ids[i - 1]
            if i != n - 1:
                tn.lower_ind_id = ind_ids[i]

        else:
            raise ValueError("Can only align vectors and operators currently.")

    if trace:
        tns[-1].lower_ind_id = tns[0].upper_ind_id

    return tns


def tensor_network_apply_op_vec(
    tn_op,
    tn_vec,
    compress=False,
    **compress_opts
):
    """Apply a general a general tensor network representing an operator (has
    ``up_ind_id`` and ``lower_ind_id``) to a tensor network representing a
    vector (has ``site_ind_id``), by contracting each pair of tensors at each
    site then compressing the resulting tensor network. How the compression
    takes place is determined by the type of tensor network passed in. The
    returned tensor network has the same site indices as ``tn_vec``, and it is
    the ``lower_ind_id`` of ``tn_op`` that is contracted.

    Parameters
    ----------
    tn_op : TensorNetwork
        The tensor network representing the operator.
    tn_vec : TensorNetwork
        The tensor network representing the vector.
    compress : bool
        Whether to compress the resulting tensor network.
    compress_opts
        Options to pass to ``tn_vec.compress``.

    Returns
    -------
    tn_op_vec : TensorNetwork
    """
    A, x = tn_op.copy(), tn_vec.copy()

    # align the indices
    coordinate_formatter = get_coordinate_formatter(A._NDIMS)
    A.lower_ind_id = f"__tmp{coordinate_formatter}__"
    A.upper_ind_id = x.site_ind_id
    x.reindex_sites_(f"__tmp{coordinate_formatter}__")

    # form total network and contract each site
    x |= A
    for site in x.gen_sites_present():
        x ^= site

    x.fuse_multibonds_()
    # optionally compress
    if compress:
        x.compress(**compress_opts)

    return x



class TensorNetworkGen(TensorNetwork):
    """A tensor network which notionally has a single tensor per 'site',
    though these could be labelled arbitrarily could also be linked in an
    arbitrary geometry by bonds.
    """

    _NDIMS = 1
    _EXTRA_PROPS = (
        "_sites",
        "_site_tag_id",
    )

    def _compatible_arbgeom(self, other):
        """Check whether ``self`` and ``other`` represent the same set of
        sites and are tagged equivalently.
        """
        return isinstance(other, TensorNetworkGen) and all(
            getattr(self, e, 0) == getattr(other, e, 1)
            for e in TensorNetworkGen._EXTRA_PROPS
        )

    def __and__(self, other):
        new = TensorNetwork.__and__(self, other)
        if self._compatible_arbgeom(other):
            new.view_as_(TensorNetworkGen, like=self)
        return new

    def __or__(self, other):
        new = TensorNetwork.__or__(self, other)
        if self._compatible_arbgeom(other):
            new.view_as_(TensorNetworkGen, like=self)
        return new

    @property
    def nsites(self):
        """The total number of sites.
        """
        return len(self._sites)

    def gen_site_coos(self):
        """Generate the coordinates of all sites, same as ``self.sites``.
        """
        return self._sites

    @property
    def sites(self):
        """Tuple of the possible sites in this tensor network.
        """
        sites = getattr(self, "_sites", None)
        if sites is None:
            sites = tuple(self.gen_site_coos())
        return sites

    def _get_site_set(self):
        """The set of all sites.
        """
        if getattr(self, "_site_set", None) is None:
            self._site_set = set(self.sites)
        return self._site_set

    def gen_sites_present(self):
        """Generate the sites which are currently present (e.g. if a local view
        of a larger tensor network), based on whether their tags are present.

        Examples
        --------

            >>> tn = qtn.TN3D_rand(4, 4, 4, 2)
            >>> tn_sub = tn.select_local('I1,2,3', max_distance=1)
            >>> list(tn_sub.gen_sites_present())
            [(0, 2, 3), (1, 1, 3), (1, 2, 2), (1, 2, 3), (1, 3, 3), (2, 2, 3)]

        """
        return (
            site for site in self.gen_site_coos()
            if self.site_tag(site) in self.tag_map
        )

    @property
    def site_tag_id(self):
        """The string specifier for tagging each site of this tensor network.
        """
        return self._site_tag_id

    def site_tag(self, site):
        """The name of the tag specifiying the tensor at ``site``.
        """
        return self.site_tag_id.format(site)

    def retag_sites(self, new_id, where=None, inplace=False):
        """Modify the site tags for all or some tensors in this tensor network
        (without changing the ``site_tag_id``).

        Parameters
        ----------
        new_id : str
            A string with a format placeholder to accept a site, e.g. "S{}".
        where : None or sequence
            Which sites to update the index labels on. If ``None`` (default)
            all sites.
        inplace : bool
            Whether to retag in place.
        """
        if where is None:
            where = self.gen_sites_present()

        return self.retag(
            {self.site_tag(x): new_id.format(x) for x in where},
            inplace=inplace,
        )

    @property
    def site_tags(self):
        """All of the site tags.
        """
        if getattr(self, "_site_tags", None) is None:
            self._site_tags = tuple(map(self.site_tag, self.gen_site_coos()))
        return self._site_tags

    @property
    def site_tags_present(self):
        """All of the site tags still present in this tensor network.
        """
        return tuple(map(self.site_tag, self.gen_sites_present()))

    @site_tag_id.setter
    def site_tag_id(self, new_id):
        if self._site_tag_id != new_id:
            self.retag_sites(new_id, inplace=True)
            self._site_tag_id = new_id
            self._site_tags = None

    def retag_all(self, new_id, inplace=False):
        """Retag all sites and change the ``site_tag_id``.
        """
        tn = self if inplace else self.copy()
        tn.site_tag_id = new_id
        return tn

    retag_all_ = functools.partialmethod(retag_all, inplace=True)

    def _get_site_tag_set(self):
        """The oset of all site tags.
        """
        if getattr(self, "_site_tag_set", None) is None:
            self._site_tag_set = set(self.site_tags)
        return self._site_tag_set

    def filter_valid_site_tags(self, tags):
        """Get the valid site tags from ``tags``.
        """
        return oset(sorted(self._get_site_tag_set().intersection(tags)))

    def maybe_convert_coo(self, x):
        """Check if ``x`` is a valid site and convert to the corresponding site
        tag if so, else return ``x``.
        """
        try:
            if x in self._get_site_set():
                return self.site_tag(x)
        except TypeError:
            pass
        return x

    def gen_tags_from_coos(self, coos):
        """Generate the site tags corresponding to the given coordinates.
        """
        return map(self.site_tag, coos)

    def _get_tids_from_tags(self, tags, which="all"):
        """This is the function that lets coordinates such as ``site`` be
        used for many 'tag' based functions.
        """
        tags = self.maybe_convert_coo(tags)
        return super()._get_tids_from_tags(tags, which=which)

    @functools.wraps(tensor_network_align)
    def align(self, *args, inplace=False, **kwargs):
        return tensor_network_align(self, *args, inplace=inplace, **kwargs)

    align_ = functools.partialmethod(align, inplace=True)


def gauge_product_boundary_vector(
    tn,
    tags,
    which="all",
    max_bond=1,
    smudge=1e-6,
    canonize_distance=0,
    select_local_distance=None,
    select_local_opts=None,
    **contract_around_opts,
):
    tids = tn._get_tids_from_tags(tags, which)

    # form the double layer tensor network - this is the TN we will
    #     generate the actual gauges with
    if select_local_distance is None:
        # use the whole tensor network ...
        outer_inds = tn.outer_inds()
        dtn = tn.H & tn
    else:
        # ... or just a local patch
        select_local_opts = ensure_dict(select_local_opts)
        ltn = tn._select_local_tids(
            tids,
            max_distance=select_local_distance,
            virtual=False,
            **select_local_opts,
        )
        outer_inds = ltn.outer_inds()
        dtn = ltn.H | ltn

    # get all inds in the tagged region
    region_inds = set.union(*(set(tn.tensor_map[tid].inds) for tid in tids))

    # contract all 'physical' indices so that we have a single layer TN
    #     outside region and double layer sandwich inside region
    for ix in outer_inds:
        if (ix in region_inds) or (ix not in dtn.ind_map):
            # 1st condition - don't contract region sandwich
            # 2nd condition - if local selecting, will get multibonds so
            #     some indices already contracted
            continue
        dtn.contract_ind(ix)

    # form the single layer boundary of double layer tagged region
    dtids = dtn._get_tids_from_tags(tags, which)
    dtn._contract_around_tids(
        dtids,
        min_distance=1,
        max_bond=max_bond,
        canonize_distance=canonize_distance,
        **contract_around_opts,
    )

    # select this boundary and compress to ensure it is a product operator
    dtn = dtn._select_without_tids(dtids, virtual=True)
    dtn.compress_all_(max_bond=1)
    dtn.squeeze_()

    # each tensor in the boundary should now have exactly two inds
    #     connecting to the top and bottom of the tagged region double
    #     layer. Iterate over these, inserting the gauge into the original
    #     tensor network that would turn each of these boundary tensors
    #     into identities.
    for t in dtn:
        (ix,) = [i for i in t.inds if i in region_inds]
        _, s, VH = do("linalg.svd", t.data)
        s = s + smudge
        G = do("reshape", s ** 0.5, (-1, 1)) * VH
        Ginv = dag(VH) * do("reshape", s ** -0.5, (1, -1))

        tid_l, tid_r = sorted(tn.ind_map[ix], key=lambda tid: tid in tids)
        tn.tensor_map[tid_l].gate_(Ginv.T, ix)
        tn.tensor_map[tid_r].gate_(G, ix)

    return tn


_VALID_GATE_PROPAGATE = {'sites', 'register', False, True}
_LAZY_GATE_CONTRACT = {
    False,
    'split-gate',
    'swap-split-gate',
    'auto-split-gate',
}


class TensorNetworkGenVector(TensorNetworkGen):
    """A tensor network which notionally has a single tensor and outer index
    per 'site', though these could be labelled arbitrarily and could also be
    linked in an arbitrary geometry by bonds.
    """

    _EXTRA_PROPS = (
        "_sites",
        "_site_tag_id",
        "_site_ind_id",
    )

    @property
    def site_ind_id(self):
        """The string specifier for the physical indices.
        """
        return self._site_ind_id

    def site_ind(self, site):
        return self.site_ind_id.format(site)

    @property
    def site_inds(self):
        """Return a tuple of all site indices.
        """
        if getattr(self, "_site_inds", None) is None:
            self._site_inds = tuple(map(self.site_ind, self.gen_site_coos()))
        return self._site_inds

    @property
    def site_inds_present(self):
        """All of the site inds still present in this tensor network.
        """
        return tuple(map(self.site_ind, self.gen_sites_present()))

    def reindex_sites(self, new_id, where=None, inplace=False):
        """Modify the site indices for all or some tensors in this vector
        tensor network (without changing the ``site_ind_id``).

        Parameters
        ----------
        new_id : str
            A string with a format placeholder to accept a site, e.g. "ket{}".
        where : None or sequence
            Which sites to update the index labels on. If ``None`` (default)
            all sites.
        inplace : bool
            Whether to reindex in place.
        """
        if where is None:
            where = self.gen_sites_present()

        return self.reindex(
            {self.site_ind(x): new_id.format(x) for x in where},
            inplace=inplace,
        )

    @site_ind_id.setter
    def site_ind_id(self, new_id):
        if self._site_ind_id != new_id:
            self.reindex_sites_(new_id)
            self._site_ind_id = new_id
            self._site_inds = None

    def reindex_all(self, new_id, inplace=False):
        """Reindex all physical sites and change the ``site_ind_id``.
        """
        tn = self if inplace else self.copy()
        tn.site_ind_id = new_id
        return tn

    reindex_all_ = functools.partialmethod(reindex_all, inplace=True)

    def gen_inds_from_coos(self, coos):
        """Generate the site inds corresponding to the given coordinates.
        """
        return map(self.site_ind, coos)

    def phys_dim(self, site=None):
        """Get the physical dimension of ``site``, defaulting to the first site
        if not specified.
        """
        if site is None:
            site = next(iter(self.gen_sites_present()))
        return self.ind_size(self.site_ind(site))

    def to_dense(
        self,
        *inds_seq,
        to_qarray=True,
        to_ket=None,
        **contract_opts
    ):
        """Contract this tensor network 'vector' into a dense array. By
        default, turn into a 'ket' ``qarray``, i.e. column vector of shape
        ``(d, 1)``.

        Parameters
        ----------
        inds_seq : sequence of sequences of str
            How to group the site indices into the dense array. By default,
            use a single group ordered like ``sites``, but only containing
            those sites which are still present.
        to_qarray : bool
            Whether to turn the dense array into a ``qarray``, if the backend
            would otherwise be ``'numpy'``.
        to_ket : None or str
            Whether to reshape the dense array into a ket (shape ``(d, 1)``
            array). If ``None`` (default), do this only if the ``inds_seq`` is
            not supplied.
        contract_opts
            Options to pass to
            :meth:`~quimb.tensor.tensor_core.TensorNewtork.contract`.

        Returns
        -------
        array
        """
        if not inds_seq:
            inds_seq = (self.site_inds_present,)
            if to_ket is None:
                to_ket = True

        x = TensorNetwork.to_dense(
            self, *inds_seq,
            to_qarray=to_qarray,
            **contract_opts
        )

        if to_ket:
            x = do("reshape", x, (-1, 1))

        return x

    def gate(
        self, G, where,
        contract=False,
        tags=None,
        propagate_tags=False,
        info=None,
        inplace=False,
        **compress_opts,
    ):
        r"""Apply a gate to this vector tensor network at sites ``where``. This
        is essentially a wrapper around
        :meth:`~quimb.tensor.tensor_core.TensorNetwork.gate_inds` apart from
        ``where`` can be specified as a list of sites, and tags can be
        optionally, intelligently propgated to the new gate tensor.

        .. math::

            | \psi \rangle \rightarrow G_\mathrm{where} | \psi \rangle

        Parameters
        ----------
        G : array_ike
            The gate array to apply, should match or be factorable into the
            shape ``(*phys_dims, *phys_dims)``.
        where : node or sequence[node]
            The sites to apply the gate to.
        contract : {False, True, 'split', 'reduce-split', 'split-gate',
                    'swap-split-gate', 'auto-split-gate'}, optional
            How to apply the gate, see
            :meth:`~quimb.tensor.tensor_core.TensorNetwork.gate_inds`.
        tags : str or sequence of str, optional
            Tags to add to the new gate tensor.
        propagate_tags : {False, True, 'register', 'sites'}, optional
            Whether to propagate tags to the new gate tensor::

                - False: no tags are propagated
                - True: all tags are propagated
                - 'register': only site tags corresponding to ``where`` are
                  added.
                - 'sites': all site tags on the current sites are propgated,
                  resulting in a lightcone like tagging.

        info : None or dict, optional
            Used to store extra optional information such as the singular
            values if not absorbed.
        inplace : bool, optional
            Whether to perform the gate operation inplace on the tensor network
            or not.
        compress_opts
            Supplied to :func:`~quimb.tensor.tensor_core.tensor_split` for any
            ``contract`` methods that involve splitting. Ignored otherwise.

        Returns
        -------
        TensorNetworkGenVector

        See Also
        --------
        TensorNetwork.gate_inds
        """
        check_opt('propagate_tags', propagate_tags, _VALID_GATE_PROPAGATE)

        tn = self if inplace else self.copy()

        if not isinstance(where, (tuple, list)):
            where = (where,)
        inds = tuple(map(tn.site_ind, where))

        # potentially add tags from current tensors to the new ones,
        # only do this if we are lazily adding the gate tensor(s)
        if (
            (contract in _LAZY_GATE_CONTRACT) and
            (propagate_tags in (True, 'sites'))
        ):
            old_tags = oset.union(
                *(t.tags for t in tn._inds_get(*inds))
            )
            if propagate_tags == 'sites':
                old_tags = tn.filter_valid_site_tags(old_tags)

            tags = tags_to_oset(tags)
            tags.update(old_tags)

        # perform the actual gating
        tn.gate_inds_(
            G, inds, contract=contract, tags=tags, info=info, **compress_opts
        )

        # possibly add tags based on where the gate was applied
        if propagate_tags == 'register':
            for ix, site in zip(inds, where):
                t, = tn._inds_get(ix)
                t.add_tag(tn.site_tag(site))

        return tn

    gate_ = functools.partialmethod(gate, inplace=True)

    def gate_simple_(self, G, where, gauges, renorm=True, **gate_opts):
        """Apply a gate to this vector tensor network at sites ``where``, using
        simple update style gauging of the tensors first, as supplied in
        ``gauges``. The new singular values for the bond are reinserted into
        ``gauges``.

        Parameters
        ----------
        G : array_like
            The gate to be applied.
        where : node or sequence[node]
            The sites to apply the gate to.
        gauges : dict[str, array_like]
            The store of gauge bonds, the keys being indices and the values
            being the vectors. Only bonds present in this dictionary will be
            used.
        renorm : bool, optional
            Whether to renormalise the singular after the gate is applied,
            before reinserting them into ``gauges``.
        """
        gate_opts.setdefault("absorb", None)
        gate_opts.setdefault("contract", "reduce-split")

        site_tags = tuple(map(self.site_tag, where))
        tn_where = self.select_any(site_tags)

        with tn_where.gauge_simple_temp(gauges, ungauge_inner=False):
            info = {}
            tn_where.gate_(G, where, info=info, **gate_opts)

            # inner ungauging is performed by tracking the new singular values
            (((_, ix), s),) = info.items()
            if renorm:
                s = s / s[0]
            gauges[ix] = s

        return self

    def gate_fit_local_(
        self,
        G,
        where,
        max_distance=0,
        fillin=0,
        gauges=None,
        **fit_opts,
    ):
        # select a local neighborhood of tensors
        tids = self._get_tids_from_tags(
            tuple(map(self.site_tag, where)), 'any'
        )
        if len(tids) == 2:
            tids = self._get_string_between_tids(*tids)

        k = self._select_local_tids(
            tids,
            max_distance=max_distance,
            fillin=fillin,
            virtual=True,
        )

        if gauges:
            outer, inner = k.gauge_simple_insert(gauges)

        Gk = k.gate(G, where)
        k.fit_(Gk, **fit_opts)

        if gauges:
            k.gauge_simple_remove(outer, inner)

        if gauges is not None:
            k.gauge_all_simple_(gauges=gauges)

    def local_expectation_cluster(
        self,
        G,
        where,
        normalized=True,
        max_distance=0,
        fillin=False,
        gauges=None,
        optimize="auto",
        max_bond=None,
        rehearse=False,
        **contract_opts,
    ):
        r"""Approximately compute a single local expectation value of the gate
        ``G`` at sites ``where``, either treating the environment beyond
        ``max_distance`` as the identity, or using simple update style bond
        gauges as supplied in ``gauges``.

        This selects a local neighbourhood of tensors up to distance
        ``max_distance`` away from ``where``, then traces over dangling bonds
        after potentially inserting the bond gauges, to form an approximate
        version of the reduced density matrix.

        .. math::

            \langle \psi | G | \psi \rangle
            \approx
            \frac{
            \mathrm{Tr} [ G \tilde{\rho}_\mathrm{where} ]
            }{
            \mathrm{Tr} [ \tilde{\rho}_\mathrm{where} ]
            }

        assuming ``normalized==True``.

        Parameters
        ----------
        G : array_like
            The gate to compute the expecation of.
        where : node or sequence[node]
            The sites to compute the expectation at.
        normalized : bool, optional
            Whether to locally normalize the result, i.e. divide by the
            expectation value of the identity.
        max_distance : int, optional
            The maximum graph distance to include tensors neighboring ``where``
            when computing the expectation. The default 0 means only the
            tensors at sites ``where`` are used.
        fillin : bool or int, optional
            When selecting the local tensors, whether and how many times to
            'fill-in' corner tensors attached multiple times to the local
            region. On a lattice this fills in the corners. See
            :meth:`~quimb.tensor.tensor_core.TensorNetwork.select_local`.
        gauges : dict[str, array_like], optional
            The store of gauge bonds, the keys being indices and the values
            being the vectors. Only bonds present in this dictionary will be
            used.
        optimize : str or PathOptimizer, optional
            The contraction path optimizer to use, when exactly contracting the
            local tensors.
        max_bond : None or int, optional
            If specified, use compressed contraction.
        rehearse : {False, 'tn', 'tree', True}, optional
            Whether to perform the computations or not::

                - False: perform the computation.
                - 'tn': return the tensor networks of each local expectation,
                  without running the path optimizer.
                - 'tree': run the path optimizer and return the
                  ``cotengra.ContractonTree`` for each local expectation.
                - True: run the path optimizer and return the ``PathInfo`` for
                  each local expectation.

        Returns
        -------
        expectation : float
        """
        # select a local neighborhood of tensors
        tids = self._get_tids_from_tags(
            tuple(map(self.site_tag, where)), 'any'
        )

        if len(tids) == 2:
            tids = self._get_string_between_tids(*tids)

        k = self._select_local_tids(
            tids,
            max_distance=max_distance,
            fillin=fillin,
            virtual=False,
        )

        if gauges is not None:
            # gauge the region with simple update style bond gauges
            k.gauge_simple_insert(gauges)

        if max_bond is not None:
            return k.local_expectation(
                G=G,
                where=where,
                max_bond=max_bond,
                optimize=optimize,
                normalized=normalized,
                rehearse=rehearse,
                **contract_opts
            )

        return k.local_expectation_exact(
            G=G,
            where=where,
            optimize=optimize,
            normalized=normalized,
            rehearse=rehearse,
            **contract_opts
        )

    local_expectation_simple = deprecated(
        local_expectation_cluster,
        'local_expectation_simple',
        'local_expectation_cluster',
    )

    def compute_local_expectation_cluster(
        self,
        terms,
        *,
        max_distance=0,
        fillin=False,
        normalized=True,
        gauges=None,
        optimize="auto",
        max_bond=None,
        return_all=False,
        rehearse=False,
        executor=None,
        progbar=False,
        **contract_opts,
    ):
        r"""Compute all local expectations of the given terms, either treating
        the environment beyond ``max_distance`` as the identity, or using
        simple update style bond gauges as supplied in ``gauges``.

        This selects a local neighbourhood of tensors up to distance
        ``max_distance`` away from each term's sites, then traces over
        dangling bonds after potentially inserting the bond gauges, to form
        an approximate version of the reduced density matrix.

        .. math::

            \sum_\mathrm{i}
            \langle \psi | G_\mathrm{i} | \psi \rangle
            \approx
            \sum_\mathrm{i}
            \frac{
            \mathrm{Tr} [ G_\mathrm{i} \tilde{\rho}_\mathrm{i} ]
            }{
            \mathrm{Tr} [ \tilde{\rho}_\mathrm{i} ]
            }

        assuming ``normalized==True``.

        Parameters
        ----------
        terms : dict[node or (node, node), array_like]
            The terms to compute the expectation of, with keys being the sites
            and values being the local operators.
        max_distance : int, optional
            The maximum graph distance to include tensors neighboring each
            term's sites when computing the expectation. The default 0 means
            only the tensors at sites of each term are used.
        fillin : bool or int, optional
            When selecting the local tensors, whether and how many times to
            'fill-in' corner tensors attached multiple times to the local
            region. On a lattice this fills in the corners. See
            :meth:`~quimb.tensor.tensor_core.TensorNetwork.select_local`.
        normalized : bool, optional
            Whether to locally normalize the result, i.e. divide by the
            expectation value of the identity. This implies that a different
            normalization factor is used for each term.
        gauges : dict[str, array_like], optional
            The store of gauge bonds, the keys being indices and the values
            being the vectors. Only bonds present in this dictionary will be
            used.
        optimize : str or PathOptimizer, optional
            The contraction path optimizer to use, when exactly contracting the
            local tensors.
        max_bond : None or int, optional
            If specified, use compressed contraction.
        return_all : bool, optional
            Whether to return all results, or just the summed expectation.
        rehearse : {False, 'tn', 'tree', True}, optional
            Whether to perform the computations or not::

                - False: perform the computation.
                - 'tn': return the tensor networks of each local expectation,
                  without running the path optimizer.
                - 'tree': run the path optimizer and return the
                  ``cotengra.ContractonTree`` for each local expectation.
                - True: run the path optimizer and return the ``PathInfo`` for
                  each local expectation.

        executor : Executor, optional
            If supplied compute the terms in parallel using this executor.
        progbar : bool, optional
            Whether to show a progress bar.
        contract_opts
            Supplied to
            :meth:`~quimb.tensor.tensor_core.TensorNetwork.contract`.

        Returns
        -------
        expecs : float or dict[node or (node, node), float]
            If ``return_all==False``, return the summed expectation value of
            the given terms. Otherwise, return a dictionary mapping each term's
            location to the expectation value.
        """
        return _compute_expecs_maybe_in_parallel(
            fn=_tn_local_expectation_cluster,
            tn=self,
            terms=terms,
            return_all=return_all,
            executor=executor,
            progbar=progbar,
            normalized=normalized,
            max_distance=max_distance,
            fillin=fillin,
            gauges=gauges,
            optimize=optimize,
            rehearse=rehearse,
            max_bond=max_bond,
            **contract_opts,
        )

    compute_local_expectation_simple = deprecated(
        compute_local_expectation_cluster,
        'compute_local_expectation_simple',
        'compute_local_expectation_cluster',
    )

    def local_expectation_exact(
        self,
        G,
        where,
        optimize="auto-hq",
        normalized=True,
        rehearse=False,
        **contract_opts,
    ):
        """Compute the local expectation of operator ``G`` at site(s) ``where``
        by exactly contracting the full overlap tensor network.
        """
        k_inds = tuple(map(self.site_ind, where))
        b_inds = tuple(map("_bra{}".format, where))
        b = self.conj().reindex_(dict(zip(k_inds, b_inds)))
        tn = (b | self)

        if rehearse:
            if rehearse == 'tn':
                return tn
            if rehearse == 'tree':
                return tn.contraction_tree(
                    optimize, output_inds=k_inds + b_inds)
            if rehearse:
                return tn.contraction_info(
                    optimize, output_inds=k_inds + b_inds)

        rho = tn.to_dense(k_inds, b_inds, optimize=optimize, **contract_opts)
        expec = do("trace", rho @ G)
        if normalized:
            expec = expec / do("trace", rho)

        return expec

    def compute_local_expectation_exact(
        self,
        terms,
        optimize="auto-hq",
        *,
        normalized=True,
        return_all=False,
        rehearse=False,
        executor=None,
        progbar=False,
        **contract_opts,
    ):
        """Compute the local expectations of many operators,
        by exactly contracting the full overlap tensor network.

        Parameters
        ----------
        terms : dict[node or (node, node), array_like]
            The terms to compute the expectation of, with keys being the sites
            and values being the local operators.
        optimize : str or PathOptimizer, optional
            The contraction path optimizer to use, when exactly contracting the
            full tensor network.
        normalized : bool, optional
            Whether to normalize the result.
        return_all : bool, optional
            Whether to return all results, or just the summed expectation.
        rehearse : {False, 'tn', 'tree', True}, optional
            Whether to perform the computations or not::

                - False: perform the computation.
                - 'tn': return the tensor networks of each local expectation,
                  without running the path optimizer.
                - 'tree': run the path optimizer and return the
                  ``cotengra.ContractonTree`` for each local expectation.
                - True: run the path optimizer and return the ``PathInfo`` for
                  each local expectation.

        executor : Executor, optional
            If supplied compute the terms in parallel using this executor.
        progbar : bool, optional
            Whether to show a progress bar.
        contract_opts
            Supplied to
            :meth:`~quimb.tensor.tensor_core.TensorNetwork.contract`.

        Returns
        -------
        expecs : float or dict[node or (node, node), float]
            If ``return_all==False``, return the summed expectation value of
            the given terms. Otherwise, return a dictionary mapping each term's
            location to the expectation value.
        """
        return _compute_expecs_maybe_in_parallel(
            fn=_tn_local_expectation_exact,
            tn=self,
            terms=terms,
            return_all=return_all,
            executor=executor,
            progbar=progbar,
            optimize=optimize,
            normalized=normalized,
            rehearse=rehearse,
            **contract_opts,
        )

    def partial_trace(
        self,
        keep,
        max_bond,
        optimize,
        flatten=True,
        reduce=False,
        normalized=True,
        symmetrized='auto',
        rehearse=False,
        method='contract_compressed',
        **contract_compressed_opts,
    ):
        """Partially trace this tensor network state, keeping only the sites in
        ``keep``, using compressed contraction.

        Parameters
        ----------
        keep : iterable of hashable
            The sites to keep.
        max_bond : int
            The maximum bond dimensions to use while compressed contracting.
        optimize : str or PathOptimizer, optional
            The contraction path optimizer to use, should specifically generate
            contractions paths designed for compressed contraction.
        flatten : {False, True, 'all'}, optional
            Whether to force 'flattening' (contracting all physical indices) of
            the tensor network before  contraction, whilst this makes the TN
            generally more complex to contract, the accuracy is usually
            improved. If ``'all'`` also flatten the tensors in ``keep``.
        reduce : bool, optional
            Whether to first 'pull' the physical indices off their respective
            tensors using QR reduction. Experimental.
        normalized : bool, optional
            Whether to normalize the reduced density matrix at the end.
        symmetrized : {'auto', True, False}, optional
            Whether to symmetrize the reduced density matrix at the end. This
            should be unecessary if ``flatten`` is set to ``True``.
        rehearse : {False, 'tn', 'tree', True}, optional
            Whether to perform the computation or not::

                - False: perform the computation.
                - 'tn': return the tensor network without running the path
                  optimizer.
                - 'tree': run the path optimizer and return the
                  ``cotengra.ContractonTree``..
                - True: run the path optimizer and return the ``PathInfo``.

        contract_compressed_opts : dict, optional
            Additional keyword arguments to pass to
            :meth:`~quimb.tensor.tensor_core.TensorNetwork.contract_compressed`.

        Returns
        -------
        rho : array_like
            The reduce density matrix of sites in ``keep``.
        """
        if symmetrized == 'auto':
            symmetrized = not flatten

        # form the partial trace
        k_inds = tuple(map(self.site_ind, keep))

        k = self.copy()
        if reduce:
            k.reduce_inds_onto_bond(*k_inds, tags='__BOND__', drop_tags=True)

        b_inds = tuple(map("_bra{}".format, keep))
        b = k.conj().reindex_(dict(zip(k_inds, b_inds)))

        tn = (b | k)
        output_inds = k_inds + b_inds

        if flatten:
            for site in self.gen_site_coos():
                if (site not in keep) or (flatten == 'all'):
                    # check if site exists still to permit e.g. local methods
                    # to use this same logic
                    tag = tn.site_tag(site)
                    if tag in tn.tag_map:
                        tn ^= tag

        tn.fuse_multibonds_()

        if method == 'contract_compressed':

            if reduce:
                output_inds = None
                tn, tn_reduced = tn.partition('__BOND__', inplace=True)

            if rehearse:
                if rehearse == 'tn':
                    return tn
                if rehearse == 'tree':
                    return tn.contraction_tree(
                        optimize, output_inds=output_inds)
                if rehearse:
                    return tn.contraction_info(
                        optimize, output_inds=output_inds)

            t_rho = tn.contract_compressed(
                optimize,
                max_bond=max_bond,
                output_inds=output_inds,
                **contract_compressed_opts,
            )

            if reduce:
                t_rho |= tn_reduced

            rho = t_rho.to_dense(k_inds, b_inds)

        elif method == 'contract_around':
            tn.contract_around_(
                tuple(map(self.site_tag, keep)), 'any',
                max_bond=max_bond,
                **contract_compressed_opts
            )

            if rehearse:
                if rehearse == 'tn':
                    return tn
                if rehearse == 'tree':
                    return tn.contraction_tree(
                        optimize, output_inds=output_inds)
                if rehearse:
                    return tn.contraction_info(
                        optimize, output_inds=output_inds)

            rho = tn.to_dense(
                k_inds, b_inds, optimize=optimize,
            )

        else:
            raise ValueError(f"Unknown method: {method}.")

        if symmetrized:
            rho = (rho + dag(rho)) / 2

        if normalized:
            rho = rho / do("trace", rho)

        return rho

    def local_expectation(
        self,
        G,
        where,
        max_bond,
        optimize,
        flatten=True,
        normalized=True,
        symmetrized='auto',
        reduce=False,
        rehearse=False,
        **contract_compressed_opts,
    ):
        """Compute the local expectation of operator ``G`` at site(s) ``where``
        by approximately contracting the full overlap tensor network.

        Parameters
        ----------
        G : array_like
            The local operator to compute the expectation of.
        where : node or sequence of nodes
            The sites to compute the expectation for.
        max_bond : int
            The maximum bond dimensions to use while compressed contracting.
        optimize : str or PathOptimizer, optional
            The contraction path optimizer to use, should specifically generate
            contractions paths designed for compressed contraction.
        method : {'rho', 'rho-reduced'}, optional
            The method to use to compute the expectation value.
        flatten : bool, optional
            Whether to force 'flattening' (contracting all physical indices) of
            the tensor network before  contraction, whilst this makes the TN
            generally more complex to contract, the accuracy is usually much
            improved.
        normalized : bool, optional
            If computing via `partial_trace`, whether to normalize the reduced
            density matrix at the end.
        symmetrized : {'auto', True, False}, optional
            If computing via `partial_trace`, whether to symmetrize the reduced
            density matrix at the end. This should be unecessary if ``flatten``
            is set to ``True``.
        rehearse : {False, 'tn', 'tree', True}, optional
            Whether to perform the computation or not::

                - False: perform the computation.
                - 'tn': return the tensor network without running the path
                  optimizer.
                - 'tree': run the path optimizer and return the
                  ``cotengra.ContractonTree``..
                - True: run the path optimizer and return the ``PathInfo``.

        contract_compressed_opts : dict, optional
            Additional keyword arguments to pass to
            :meth:`~quimb.tensor.tensor_core.TensorNetwork.contract_compressed`.

        Returns
        -------
        expec : float
        """
        rho = self.partial_trace(
            keep=where,
            max_bond=max_bond,
            optimize=optimize,
            flatten=flatten,
            reduce=reduce,
            normalized=normalized,
            symmetrized=symmetrized,
            rehearse=rehearse,
            **contract_compressed_opts,
        )
        if rehearse:
            return rho

        return do("trace", rho @ G)

    def compute_local_expectation(
        self,
        terms,
        max_bond,
        optimize,
        *,
        flatten=True,
        normalized=True,
        symmetrized='auto',
        reduce=False,
        return_all=False,
        rehearse=False,
        executor=None,
        progbar=False,
        **contract_compressed_opts,
    ):
        """Compute the local expectations of many local operators, by
        approximately contracting the full overlap tensor network.

        Parameters
        ----------
        terms : dict[node or (node, node), array_like]
            The terms to compute the expectation of, with keys being the sites
            and values being the local operators.
        max_bond : int
            The maximum bond dimension to use during contraction.
        optimize : str or PathOptimizer
            The compressed contraction path optimizer to use.
        method : {'rho', 'rho-reduced'}, optional
            The method to use to compute the expectation value.

                - 'rho': compute the expectation value via the reduced density
                  matrix.
                - 'rho-reduced': compute the expectation value via the reduced
                  density matrix, having reduced the physical indices onto the
                  bonds first.

        flatten : bool, optional
            Whether to force 'flattening' (contracting all physical indices) of
            the tensor network before  contraction, whilst this makes the TN
            generally more complex to contract, the accuracy can often be much
            improved.
        normalized : bool, optional
            Whether to locally normalize the result.
        symmetrized : {'auto', True, False}, optional
            Whether to symmetrize the reduced density matrix at the end. This
            should be unecessary if ``flatten`` is set to ``True``.
        return_all : bool, optional
            Whether to return all results, or just the summed expectation. If
            ``rehease is not False``, this is ignored and a dict is always
            returned.
        rehearse : {False, 'tn', 'tree', True}, optional
            Whether to perform the computations or not::

                - False: perform the computation.
                - 'tn': return the tensor networks of each local expectation,
                  without running the path optimizer.
                - 'tree': run the path optimizer and return the
                  ``cotengra.ContractonTree`` for each local expectation.
                - True: run the path optimizer and return the ``PathInfo`` for
                  each local expectation.

        executor : Executor, optional
            If supplied compute the terms in parallel using this executor.
        progbar : bool, optional
            Whether to show a progress bar.
        contract_compressed_opts
            Supplied to
            :meth:`~quimb.tensor.tensor_core.TensorNetwork.contract_compressed`.

        Returns
        -------
        expecs : float or dict[node or (node, node), float]
            If ``return_all==False``, return the summed expectation value of
            the given terms. Otherwise, return a dictionary mapping each term's
            location to the expectation value.
        """
        return _compute_expecs_maybe_in_parallel(
            fn=_tn_local_expectation,
            tn=self,
            terms=terms,
            return_all=return_all,
            executor=executor,
            progbar=progbar,
            max_bond=max_bond,
            optimize=optimize,
            normalized=normalized,
            symmetrized=symmetrized,
            reduce=reduce,
            flatten=flatten,
            rehearse=rehearse,
            **contract_compressed_opts,
        )

    compute_local_expectation_rehearse = functools.partialmethod(
        compute_local_expectation, rehearse=True)

    compute_local_expectation_tn = functools.partialmethod(
        compute_local_expectation, rehearse='tn')


class TensorNetworkGenOperator(TensorNetworkGen):
    """A tensor network which notionally has a single tensor and two outer
    indices per 'site', though these could be labelled arbitrarily and could
    also be linked in an arbitrary geometry by bonds.
    """

    _EXTRA_PROPS = (
        "_sites",
        "_site_tag_id",
        "_upper_ind_id",
        "_lower_ind_id",
    )

    @property
    def upper_ind_id(self):
        """The string specifier for the upper phyiscal indices.
        """
        return self._upper_ind_id

    @upper_ind_id.setter
    def upper_ind_id(self, new_id):
        if new_id == self._lower_ind_id:
            raise ValueError("Setting the same upper and upper index ids will"
                             " make the two ambiguous.")

        if self._upper_ind_id != new_id:
            self.reindex_upper_sites_(new_id)
            self._upper_ind_id = new_id
            self._upper_inds = None

    def upper_ind(self, site):
        """Get the upper physical index name of ``site``.
        """
        return self.upper_ind_id.format(site)

    @property
    def upper_inds(self):
        """Return a tuple of all upper indices.
        """
        if getattr(self, "_upper_inds", None) is None:
            self._upper_inds = tuple(map(self.upper_ind, self.gen_site_coos()))
        return self._upper_inds

    @property
    def upper_inds_present(self):
        """Return a tuple of all upper indices still present in the tensor
        network.
        """
        return tuple(map(self.upper_ind, self.gen_sites_present()))

    @property
    def lower_ind_id(self):
        """The string specifier for the lower phyiscal indices.
        """
        return self._lower_ind_id

    @lower_ind_id.setter
    def lower_ind_id(self, new_id):
        if new_id == self._upper_ind_id:
            raise ValueError("Setting the same upper and lower index ids will"
                             " make the two ambiguous.")

        if self._lower_ind_id != new_id:
            self.reindex_lower_sites_(new_id)
            self._lower_ind_id = new_id
            self._lower_inds = None

    def lower_ind(self, site):
        """Get the lower physical index name of ``site``.
        """
        return self.lower_ind_id.format(site)

    @property
    def lower_inds(self):
        """Return a tuple of all lower indices.
        """
        if getattr(self, "_lower_inds", None) is None:
            self._lower_inds = tuple(map(self.lower_ind, self.gen_site_coos()))
        return self._lower_inds

    @property
    def lower_inds_present(self):
        """Return a tuple of all lower indices still present in the tensor
        network.
        """
        return tuple(map(self.lower_ind, self.gen_sites_present()))

    def to_dense(
        self,
        *inds_seq,
        to_qarray=True,
        **contract_opts
    ):
        """Contract this tensor network 'operator' into a dense array.

        Parameters
        ----------
        inds_seq : sequence of sequences of str
            How to group the site indices into the dense array. By default,
            use a single group ordered like ``sites``, but only containing
            those sites which are still present.
        to_qarray : bool
            Whether to turn the dense array into a ``qarray``, if the backend
            would otherwise be ``'numpy'``.
        contract_opts
            Options to pass to
            :meth:`~quimb.tensor.tensor_core.TensorNewtork.contract`.

        Returns
        -------
        array
        """
        if not inds_seq:
            inds_seq = (self.upper_inds_present, self.lower_inds_present)

        return TensorNetwork.to_dense(
            self, *inds_seq,
            to_qarray=to_qarray,
            **contract_opts
        )

    def phys_dim(self, site=None, which='upper'):
        """Get the physical dimension of ``site``.
        """
        if site is None:
            site = next(iter(self.gen_sites_present()))

        if which == 'upper':
            return self[site].ind_size(self.upper_ind(site))

        if which == 'lower':
            return self[site].ind_size(self.lower_ind(site))


def _compute_expecs_maybe_in_parallel(
    fn,
    tn,
    terms,
    return_all=False,
    executor=None,
    progbar=False,
    **kwargs,
):
    """Unified helper function for the various methods that compute many
    expectations, possibly in parallel, possibly with a progress bar.
    """
    if not isinstance(terms, dict):
        terms = dict(terms.items())

    if executor is None:
        results = (fn(tn, G, where, **kwargs) for where, G in terms.items())
    else:
        if hasattr(executor, 'scatter'):
            tn = executor.scatter(tn)

        futures = [
            executor.submit(fn, tn, G, where, **kwargs)
            for where, G in terms.items()
        ]
        results = (future.result() for future in futures)

    if progbar:
        results = Progbar(results, total=len(terms))

    expecs = dict(zip(terms.keys(), results))

    if return_all or kwargs.get('rehearse', False):
        return expecs

    return functools.reduce(add, expecs.values())


def _tn_local_expectation(tn, *args, **kwargs):
    """Define as function for pickleability.
    """
    return tn.local_expectation(*args, **kwargs)


def _tn_local_expectation_cluster(tn, *args, **kwargs):
    """Define as function for pickleability.
    """
    return tn.local_expectation_cluster(*args, **kwargs)


def _tn_local_expectation_exact(tn, *args, **kwargs):
    """Define as function for pickleability.
    """
    return tn.local_expectation_exact(*args, **kwargs)




"""Approximating spectral functions with tensor networks.
"""

import numpy as np
import random

import quimb as qu

from .tensor_builder import MPO_rand, MPO_zeros_like


def construct_lanczos_tridiag_MPO(A, K, v0=None, initial_bond_dim=None,
                                  beta_tol=1e-6, max_bond=None, seed=False,
                                  v0_opts=None, k_min=10):
    """
    """
    if initial_bond_dim is None:
        initial_bond_dim = 8
    if max_bond is None:
        max_bond = 8

    if v0 is None:
        if seed:
            # needs to be truly random so MPI processes don't overlap
            qu.seed_rand(random.SystemRandom().randint(0, 2**32 - 1))

        V = MPO_rand(A.L, initial_bond_dim,
                     phys_dim=A.phys_dim(), dtype=A.dtype)
    else:  # normalize
        V = v0 / (v0.H @ v0)**0.5
    Vm1 = MPO_zeros_like(V)

    alpha = np.zeros(K + 1)
    beta = np.zeros(K + 2)

    bsz = A.phys_dim()**A.L
    beta[1] = bsz  # == sqrt(prod(A.shape))

    compress_kws = {'max_bond': max_bond, 'method': 'svd'}

    for j in range(1, K + 1):

        Vt = A.apply(V, compress=True, **compress_kws)
        Vt.add_MPO(-beta[j] * Vm1, inplace=True, compress=True, **compress_kws)
        alpha[j] = (V.H @ Vt).real
        Vt.add_MPO(-alpha[j] * V, inplace=True, compress=True, **compress_kws)
        beta[j + 1] = (Vt.H @ Vt)**0.5

        # check for convergence
        if abs(beta[j + 1]) < beta_tol:
            yield alpha[1:j + 1], beta[2:j + 2], beta[1]**2 / bsz
            break

        Vm1 = V.copy()
        V = Vt / beta[j + 1]

        if j >= k_min:
            yield (np.copy(alpha[1:j + 1]),
                   np.copy(beta[2:j + 2]),
                   np.copy(beta[1])**2 / bsz)
